# -*- coding: utf-8 -*-
"""
Unit tests for the kCSD methods

This was written by :
Chaitanya Chintaluri,
Laboratory of Neuroinformatics,
Nencki Institute of Exprimental Biology, Warsaw.

:license: Modified BSD, see LICENSE.txt for details.
"""

import unittest
import neo
import numpy as np
import quantities as pq
from elephant import current_source_density as CSD
import elephant.current_source_density_src.utility_functions as utils


class KCSD1D_TestCase(unittest.TestCase):
    def setUp(self):
        self.ele_pos = utils.generate_electrodes(dim=1).reshape(5, 1)
        self.csd_profile = utils.gauss_1d_dipole
        pots = CSD.generate_lfp(self.csd_profile, self.ele_pos)
        self.pots = np.reshape(pots, (-1, 1))
        self.test_method = 'KCSD1D'
        self.test_params = {'h': 50.}

        temp_signals = []
        for ii in range(len(self.pots)):
            temp_signals.append(self.pots[ii])
        self.an_sigs = neo.AnalogSignal(np.array(temp_signals).T * pq.mV,
                                        sampling_rate=1000 * pq.Hz)
        self.an_sigs.annotate(coordinates=self.ele_pos * pq.mm)

    def test_kcsd1d_estimate(self, cv_params={}):
        self.test_params.update(cv_params)
        result = CSD.estimate_csd(self.an_sigs, method=self.test_method,
                                  **self.test_params)
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(result.times, [0.] * pq.s)
        self.assertEqual(len(result.annotations.keys()), 1)
        true_csd = self.csd_profile(result.annotations['x_coords'])
        rms = np.linalg.norm(np.array(result[0, :]) - true_csd)
        rms /= np.linalg.norm(true_csd)
        self.assertLess(rms, 0.5, msg='RMS between trueCSD and estimate > 0.5')

    def test_valid_inputs(self):
        self.test_method = 'InvalidMethodName'
        self.assertRaises(ValueError, self.test_kcsd1d_estimate)
        self.test_method = 'KCSD1D'
        self.test_params = {'src_type': 22}
        self.assertRaises(KeyError, self.test_kcsd1d_estimate)
        self.test_method = 'KCSD1D'
        self.test_params = {'InvalidKwarg': 21}
        self.assertRaises(TypeError, self.test_kcsd1d_estimate)
        cv_params = {'InvalidCVArg': np.array((0.1, 0.25, 0.5))}
        self.assertRaises(TypeError, self.test_kcsd1d_estimate, cv_params)


class KCSD2D_TestCase(unittest.TestCase):
    def setUp(self):
        xx_ele, yy_ele = utils.generate_electrodes(dim=2, res=9,
                                                   xlims=[0.05, 0.95],
                                                   ylims=[0.05, 0.95])
        self.ele_pos = np.vstack((xx_ele, yy_ele)).T
        self.csd_profile = utils.large_source_2D
        pots = CSD.generate_lfp(
            self.csd_profile,
            xx_ele,
            yy_ele,
            resolution=100)
        self.pots = np.reshape(pots, (-1, 1))
        self.test_method = 'KCSD2D'
        self.test_params = {'gdx': 0.25, 'gdy': 0.25, 'R_init': 0.08,
                            'h': 50., 'xmin': 0., 'xmax': 1.,
                            'ymin': 0., 'ymax': 1.}
        temp_signals = []
        for ii in range(len(self.pots)):
            temp_signals.append(self.pots[ii])
        self.an_sigs = neo.AnalogSignal(np.array(temp_signals).T * pq.mV,
                                        sampling_rate=1000 * pq.Hz)
        self.an_sigs.annotate(coordinates=self.ele_pos * pq.mm)

    def test_kcsd2d_estimate(self, cv_params={}):
        self.test_params.update(cv_params)
        result = CSD.estimate_csd(self.an_sigs, method=self.test_method,
                                  **self.test_params)
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(result.times, [0.] * pq.s)
        self.assertEqual(len(result.annotations.keys()), 2)
        true_csd = self.csd_profile(result.annotations['x_coords'],
                                    result.annotations['y_coords'])
        rms = np.linalg.norm(np.array(result[0, :]) - true_csd)
        rms /= np.linalg.norm(true_csd)
        self.assertLess(rms, 0.5, msg='RMS ' + str(rms) +
                        'between trueCSD and estimate > 0.5')

    def test_moi_estimate(self):
        result = CSD.estimate_csd(self.an_sigs, method='MoIKCSD',
                                  MoI_iters=10, lambd=0.0,
                                  gdx=0.2, gdy=0.2)
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(result.times, [0.] * pq.s)
        self.assertEqual(len(result.annotations.keys()), 2)

    def test_valid_inputs(self):
        self.test_method = 'InvalidMethodName'
        self.assertRaises(ValueError, self.test_kcsd2d_estimate)
        self.test_method = 'KCSD2D'
        self.test_params = {'src_type': 22}
        self.assertRaises(KeyError, self.test_kcsd2d_estimate)
        self.test_params = {'InvalidKwarg': 21}
        self.assertRaises(TypeError, self.test_kcsd2d_estimate)
        cv_params = {'InvalidCVArg': np.array((0.1, 0.25, 0.5))}
        self.assertRaises(TypeError, self.test_kcsd2d_estimate, cv_params)


class KCSD3D_TestCase(unittest.TestCase):
    def setUp(self):
        xx_ele, yy_ele, zz_ele = utils.generate_electrodes(dim=3, res=5,
                                                           xlims=[0.15, 0.85],
                                                           ylims=[0.15, 0.85],
                                                           zlims=[0.15, 0.85])
        self.ele_pos = np.vstack((xx_ele, yy_ele, zz_ele)).T
        self.csd_profile = utils.gauss_3d_dipole
        pots = CSD.generate_lfp(self.csd_profile, xx_ele, yy_ele, zz_ele)
        self.pots = np.reshape(pots, (-1, 1))
        self.test_method = 'KCSD3D'
        self.test_params = {'gdx': 0.05, 'gdy': 0.05, 'gdz': 0.05,
                            'lambd': 5.10896977451e-19, 'src_type': 'step',
                            'R_init': 0.31, 'xmin': 0., 'xmax': 1., 'ymin': 0.,
                            'ymax': 1., 'zmin': 0., 'zmax': 1.}

        temp_signals = []
        for ii in range(len(self.pots)):
            temp_signals.append(self.pots[ii])
        self.an_sigs = neo.AnalogSignal(np.array(temp_signals).T * pq.mV,
                                        sampling_rate=1000 * pq.Hz)
        self.an_sigs.annotate(coordinates=self.ele_pos * pq.mm)

    def test_kcsd3d_estimate(self, cv_params={}):
        self.test_params.update(cv_params)
        result = CSD.estimate_csd(self.an_sigs, method=self.test_method,
                                  **self.test_params)
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(result.times, [0.] * pq.s)
        self.assertEqual(len(result.annotations.keys()), 3)
        true_csd = self.csd_profile(result.annotations['x_coords'],
                                    result.annotations['y_coords'],
                                    result.annotations['z_coords'])
        rms = np.linalg.norm(np.array(result[0, :]) - true_csd)
        rms /= np.linalg.norm(true_csd)
        self.assertLess(rms, 0.5, msg='RMS ' + str(rms) +
                        ' between trueCSD and estimate > 0.5')

    def test_valid_inputs(self):
        self.test_method = 'InvalidMethodName'
        self.assertRaises(ValueError, self.test_kcsd3d_estimate)
        self.test_method = 'KCSD3D'
        self.test_params = {'src_type': 22}
        self.assertRaises(KeyError, self.test_kcsd3d_estimate)
        self.test_params = {'InvalidKwarg': 21}
        self.assertRaises(TypeError, self.test_kcsd3d_estimate)
        cv_params = {'InvalidCVArg': np.array((0.1, 0.25, 0.5))}
        self.assertRaises(TypeError, self.test_kcsd3d_estimate, cv_params)


if __name__ == '__main__':
    unittest.main()



import hashlib
import os
import ssl
import tempfile
from urllib.parse import urlparse
import warnings
from os import environ, getenv
from pathlib import Path
from urllib.error import HTTPError, URLError
from urllib.request import urlopen, urlretrieve
from zipfile import ZipFile

from tqdm import tqdm

from elephant import _get_version

ELEPHANT_TMP_DIR = Path(tempfile.gettempdir()) / "elephant"


class TqdmUpTo(tqdm):
    """
    Provides `update_to(n)` which uses `tqdm.update(delta_n)`.
    Original implementation:
    https://github.com/tqdm/tqdm/blob/master/examples/tqdm_wget.py
    """

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b : int, optional
            Number of blocks transferred so far [default: 1].
        bsize : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def calculate_md5(filepath, chunk_size=1024 * 1024):
    md5 = hashlib.md5()
    with open(filepath, 'rb') as f:
        for chunk in iter(lambda: f.read(chunk_size), b''):
            md5.update(chunk)
    return md5.hexdigest()


def check_integrity(filepath, md5):
    if not Path(filepath).exists() or md5 is None:
        return False
    return calculate_md5(filepath) == md5


def download(url, filepath=None, checksum=None, verbose=True):
    if filepath is None:
        filename = url.split('/')[-1]
        filepath = ELEPHANT_TMP_DIR / filename
    filepath = Path(filepath)
    if check_integrity(filepath, md5=checksum):
        return filepath
    folder = filepath.absolute().parent
    folder.mkdir(exist_ok=True)
    desc = f"Downloading {url} to '{filepath}'"
    with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1,
                  desc=desc, disable=not verbose) as t:
        try:
            urlretrieve(url, filename=filepath, reporthook=t.update_to)
        except URLError:
            # do not authenticate SSL certificate
            ssl._create_default_https_context = ssl._create_unverified_context
            urlretrieve(url, filename=filepath, reporthook=t.update_to)
    return filepath


def download_datasets(repo_path, filepath=None, checksum=None,
                      verbose=True):
    r"""
    This function can be used to download files from elephant-data using
    only the path relative to the root of the elephant-data repository.
    The default URL used, points to elephants corresponding release of
    elephant-data.
    Different versions of the elephant package may require different
    versions of elephant-data.
    e.g. the following URLs:
    -  https://web.gin.g-node.org/NeuralEnsemble/elephant-data/raw/0.0.1
       points to release v0.0.1.
    -  https://web.gin.g-node.org/NeuralEnsemble/elephant-data/raw/master
       always points to the latest state of elephant-data.
    -  https://datasets.python-elephant.org/
       points to the root of elephant data

    To change this URL, use the environment variable `ELEPHANT_DATA_LOCATION`.
    When using data, which is not yet contained in the master branch or a
    release of elephant data, e.g. during development, this variable can
    be used to change the default URL.
    For example to use data on branch `multitaper`, change the
    `ELEPHANT_DATA_LOCATION` to
    https://web.gin.g-node.org/NeuralEnsemble/elephant-data/raw/multitaper.
    For a complete example, see Examples section.
        
    To use a local copy of elephant-data, use the environment variable
    `ELEPHANT_DATA_LOCATION`, e.g. set to /home/user/elephant-data.
        
    Parameters
    ----------
    repo_path : str
        String denoting the path relative to elephant-data repository root
    filepath : str, optional
        Path to temporary folder where the downloaded files will be stored
    checksum : str, optional
        Checksum to verify data integrity after download
    verbose : bool, optional
        Whether to disable the entire progressbar wrapper [].
        If set to None, disable on non-TTY.
        Default: True

    Returns
    -------
    filepath : pathlib.Path
        Path to downloaded files.


    Notes
    -----
    The default URL always points to elephant-data. Please
    do not change its value. For development purposes use the environment
    variable 'ELEPHANT_DATA_LOCATION'.

    Examples
    --------
    The following example downloads a file from elephant-data branch
    'multitaper', by setting the environment variable to the branch URL:

    >>> import os
    >>> from elephant.datasets import download_datasets
    >>> os.environ["ELEPHANT_DATA_LOCATION"] = "https://web.gin.g-node.org/NeuralEnsemble/elephant-data/raw/multitaper" # noqa
    >>> download_datasets("unittest/spectral/multitaper_psd/data/time_series.npy") # doctest: +SKIP
    PosixPath('/tmp/elephant/time_series.npy')
    """

    env_var = 'ELEPHANT_DATA_LOCATION'
    if env_var in os.environ:  # user did set path or URL
        if os.path.exists(getenv(env_var)):
            return Path(f"{getenv(env_var)}/{repo_path}")
        elif urlparse(getenv(env_var)).scheme not in ('http', 'https'):
            raise ValueError(f"The environment variable {env_var} must be set to either an existing file system path "
                             f"or a valid URL. Given value: '{getenv(env_var)}' is neither.")

    # this url redirects to the current location of elephant-data
    url_to_root = "https://datasets.python-elephant.org/"

    # get URL to corresponding version of elephant data
    # (version elephant is equal to version elephant-data)
    default_url = url_to_root + f"raw/v{_get_version()}"

    if env_var not in environ:  # user did not set URL
        # is 'version-URL' available? (not for elephant development version)
        try:
            urlopen(default_url+'/README.md')

        except HTTPError as error:
            # if corresponding elephant-data version is not found,
            # use latest commit of elephant-data
            default_url = url_to_root + "raw/master"

            warnings.warn(f"No corresponding version of elephant-data found.\n"
                          f"Elephant version: {_get_version()}. "
                          f"Data URL:{error.url}, error: {error}.\n"
                          f"Using elephant-data latest instead (This is "
                          f"expected for elephant development versions).")

        except URLError as error:
            # if verification of SSL certificate fails, do not verify cert
            try:  # try again without certificate verification
                ctx = ssl._create_unverified_context()
                ctx.check_hostname = True
                urlopen(default_url + '/README.md')
            except HTTPError:  # e.g. 404
                default_url = url_to_root + "raw/master"

            warnings.warn(f"Data URL:{default_url}, error: {error}."
                          f"{error.reason}")

    url = f"{getenv(env_var, default_url)}/{repo_path}"

    return download(url, filepath, checksum, verbose)


def unzip(filepath, outdir=ELEPHANT_TMP_DIR, verbose=True):
    with ZipFile(filepath) as zfile:
        zfile.extractall(path=outdir)
    if verbose:
        print(f"Extracted {filepath} to {outdir}")



# -*- coding: utf-8 -*-
"""
This module provides function to estimate causal influences of signals on each
other.


Granger causality
*****************
Granger causality is a method to determine causal influence of one signal on
another based on autoregressive modelling. It was developed by Nobel prize
laureate Clive Granger and has been adopted in various numerical fields ever
since :cite:`granger-Granger69_424`. In its simplest form, the
method tests whether the past values of one signal help to reduce the
prediction error of another signal, compared to the past values of the latter
signal alone. If it does reduce the prediction error, the first signal is said
to Granger cause the other signal.
Granger causality analysis can be extended to the spectral domain investigating
the influnece signals have onto each other in a frequency resolved manner. It
relies on estimating the cross-spectrum of time series and decomposing them
into a transfer function and a noise covariance matrix. The method to estimate
the spectral Granger causality is non-parametric in the sense that it does not
require to fit an autoregressive model (see :cite:`granger-Dhamala08_354`).

Limitations
-----------
The user must be mindful of the method's limitations, which are assumptions of
covariance stationary data, linearity imposed by the underlying autoregressive
modelling as well as the fact that the variables not included in the model will
not be accounted for :cite:`granger-Seth07_1667`.
When estimating spectral Granger causality the user must be familiar with
basics the multitaper method to estimate power- and cross-spectra (e.g.
sampling frequency, DPSS, time-half bandwidth product).

Implementation
--------------
The mathematical implementation of Granger causality methods in this module
closely follows :cite:`granger-Ding06_0608035`.
The implementation of spectral Granger causality follows
:cite:`granger-Dhamala08_354`, :cite:`granger-Wen13_20110610` and
:cite:`granger-Wilson72_420` for the spectral matrix factorization.


Overview of Functions
---------------------
Various formulations of Granger causality have been developed. In this module
you will find function for time-series data to test pairwise Granger causality
(`pairwise_granger`).

Time-series Granger causality
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: _toctree/causality/

    pairwise_granger
    conditional_granger

Spectral Granger causality
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: _toctree/causality/

    pairwise_spectral_granger


Tutorial
--------

:doc:`View tutorial <../tutorials/granger_causality>`

Run tutorial interactively:

.. image:: https://mybinder.org/badge.svg
   :target: https://mybinder.org/v2/gh/NeuralEnsemble/elephant/master
            ?filepath=doc/tutorials/granger_causality.ipynb

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals

import warnings
from collections import namedtuple

import numpy as np
import quantities as pq
import neo
from neo.core import AnalogSignal
from elephant.spectral import segmented_multitaper_cross_spectrum, multitaper_psd


__all__ = (
    "Causality",
    "pairwise_granger",
    "conditional_granger",
    "pairwise_spectral_granger"
)


# the return type of pairwise_granger(), pairwise_spectral_granger() function
Causality = namedtuple('Causality',
                       ['directional_causality_x_y',
                        'directional_causality_y_x',
                        'instantaneous_causality',
                        'total_interdependence'])


def _bic(cov, order, dimension, length):
    """
    Calculate Bayesian Information Criterion

    Parameters
    ----------
    cov : np.ndarray
        covariance matrix of auto regressive model
    order : int
        order of autoregressive model
    dimension : int
        dimensionality of the data
    length : int
        number of time samples

    Returns
    -------
    criterion : float
       Bayesian Information Criterion
    """
    sign, log_det_cov = np.linalg.slogdet(cov)
    criterion = 2 * log_det_cov \
        + 2*(dimension**2)*order*np.log(length)/length

    return criterion


def _aic(cov, order, dimension, length):
    """
    Calculate Akaike Information Criterion

    Parameters
    ----------
    cov : np.ndarray
        covariance matrix of auto regressive model
    order : int
        order of autoregressive model
    dimension : int
        dimensionality of the data
    length : int
        number of time samples

    Returns
    -------
    criterion : float
        Akaike Information Criterion
    """
    sign, log_det_cov = np.linalg.slogdet(cov)
    criterion = 2 * log_det_cov \
        + 2*(dimension**2)*order/length

    return criterion


def _lag_covariances(signals, dimension, max_lag):
    r"""
    Determine covariances of time series and time shift of itself up to a
    maximal lag

    Parameters
    ----------
    signals : np.ndarray
        time series data
    dimension : int
        number of time series
    max_lag : int
        maximal time lag to be considered

    Returns
    -------
    lag_corr : np.ndarray
        correlations matrices of lagged signals

    Covariance of shifted signals calculated according to the following
    formula:

        x: d-dimensional signal
        x^T: transpose of d-dimensional signal
        N: number of time points
        \tau: lag

        C(\tau) = \sum_{i=0}^{N-\tau} x[i]*x^T[\tau+i]

    """
    length = np.size(signals[0])

    if length < max_lag:
        raise ValueError("Maximum lag larger than size of data")

    # centralize time series
    signals_mean = (signals - np.mean(signals, keepdims=True)).T

    lag_covariances = np.zeros((max_lag+1, dimension, dimension))

    # determine lagged covariance for different time lags
    for lag in range(0, max_lag+1):
        lag_covariances[lag] = \
                np.mean(np.einsum('ij,ik -> ijk', signals_mean[:length-lag],
                                  signals_mean[lag:]), axis=0)

    return lag_covariances


def _yule_walker_matrix(data, dimension, order):
    r"""
    Generate matrix for Yule-Walker equation

    Parameters
    ----------
    data : np.ndarray
        correlation of data shifted with lags up to order
    dimension : int
        dimensionality of data (e.g. number of channels)
    order : int
        order of the autoregressive model

    Returns
    -------
    yule_walker_matrix : np.ndarray
        matrix in Yule-Walker equation

        Yule-Walker Matrix M is a block-structured symmetric matrix with
        dimension (d \cdot p)\times(d \cdot p)

        where
        d: dimension of signal
        p: order of autoregressive model
        C(\tau): time-shifted covariances \tau -> d \times d matrix

        The blocks of size (d \times d) are set as follows:

        M_ij = C(j-i)^T

        where 1 \leq i \leq j \leq p. The other entries are determined by
        symmetry.
    lag_covariances : np.ndarray

    """

    lag_covariances = _lag_covariances(data, dimension, order)

    yule_walker_matrix = np.zeros((dimension*order, dimension*order))

    for block_row in range(order):
        for block_column in range(block_row, order):
            yule_walker_matrix[block_row*dimension: (block_row+1)*dimension,
                               block_column*dimension:
                               (block_column+1)*dimension] = \
                lag_covariances[block_column-block_row].T

            yule_walker_matrix[block_column*dimension:
                               (block_column+1)*dimension,
                               block_row*dimension:
                               (block_row+1)*dimension] = \
                lag_covariances[block_column-block_row]
    return yule_walker_matrix, lag_covariances


def _vector_arm(signals, dimension, order):
    r"""
    Determine coefficients of autoregressive model from time series data.

    Coefficients of autoregressive model calculated via solving the linear
    equation

    M A = C

    where
    M: Yule-Waler Matrix
    A: Coefficients of autoregressive model
    C: Time-shifted covariances with positive lags

    Covariance matrix C_0 is then given by

    C_0 = C[0] - \sum_{i=0}^{p-1} A[i]C[i+1]

    where p is the orde of the autoregressive model.

    Parameters
    ----------
    signals : np.ndarray
        time series data
    order : int
        order of the autoregressive model

    Returns
    -------
    coeffs : np.ndarray
        coefficients of the autoregressive model
        ry
    covar_mat : np.ndarray
        covariance matrix of

    """

    yule_walker_matrix, lag_covariances = \
        _yule_walker_matrix(signals, dimension, order)

    positive_lag_covariances = np.reshape(lag_covariances[1:],
                                          (dimension*order, dimension))

    lstsq_coeffs = np.linalg.lstsq(yule_walker_matrix,
                                   positive_lag_covariances,
                                   rcond=None)[0]

    coeffs = []
    for index in range(order):
        coeffs.append(lstsq_coeffs[index*dimension:(index+1)*dimension, ].T)

    coeffs = np.stack(coeffs)

    cov_matrix = np.copy(lag_covariances[0])
    for i in range(order):
        cov_matrix -= np.matmul(coeffs[i], lag_covariances[i+1])

    return coeffs, cov_matrix


def _optimal_vector_arm(signals, dimension, max_order,
                        information_criterion='aic'):
    """
    Determine optimal auto regressive model by choosing optimal order via
    Information Criterion

    Parameters
    ----------
    signals : np.ndarray
        time series data
    dimension : int
        dimensionality of the data
    max_order : int
        maximal order to consider
    information_criterion : str
        A function to compute the information criterion:
            `bic` for Bayesian information_criterion,
            `aic` for Akaike information criterion
        Default: aic

    Returns
    -------
    optimal_coeffs : np.ndarray
        coefficients of the autoregressive model
    optimal_cov_mat : np.ndarray
        covariance matrix of
    optimal_order : int
        optimal order
    """

    length = np.size(signals[0])

    optimal_ic = np.infty
    optimal_order = 1
    optimal_coeffs = np.zeros((dimension, dimension, optimal_order))
    optimal_cov_matrix = np.zeros((dimension, dimension))

    for order in range(1, max_order + 1):
        coeffs, cov_matrix = _vector_arm(signals, dimension, order)

        if information_criterion == 'aic':
            temp_ic = _aic(cov_matrix, order, dimension, length)
        elif information_criterion == 'bic':
            temp_ic = _bic(cov_matrix, order, dimension, length)
        else:
            raise ValueError("The specified information criterion is not"
                             "available. Please use 'aic' or 'bic'.")

        if temp_ic < optimal_ic:
            optimal_ic = temp_ic
            optimal_order = order
            optimal_coeffs = coeffs
            optimal_cov_matrix = cov_matrix

    return optimal_coeffs, optimal_cov_matrix, optimal_order


def _bracket_operator(spectrum, num_freqs, num_signals):
    r"""
    Implementation of the $[ \cdot ]^{+}$ from 'The Factorization of Matricial
    Spectral Densities', Wilson 1972, SiAM J Appl Math, Definition 1.2 (ii).

    Paramaters
    ----------
    spectrum : np.ndarray
        Cross-spectrum of multivariate time series
    num_freqs : int
        Number of frequencies
    num_signals : int
        Number of time series

    Returns
    -------
    causal_part : np.ndarray
        Causal part of cross-spectrum of multivariate time series
    """
    # Get coefficients from spectrum
    causal_part = np.fft.ifft(spectrum, axis=0)

    # Throw away acausal part
    causal_part[(num_freqs + 1) // 2:] = 0

    # Treat coefficient belonging to 0
    causal_part[0] /= 2

    # Back-transformation
    causal_part = np.fft.fft(causal_part, axis=0)

    # Adjust zero frequency part to ensure convergence by setting entries
    # below diagonal to zero at zero-frequency
    if num_signals > 1:
        indices = np.tril_indices(num_signals, k=-1)
        causal_part[0, indices[0], indices[1]] = 0

    return causal_part


def _dagger(matrix_array):
    r"""
    Return Hermitian conjugate of matrix array.

    Parameters
    ----------
    matrix_array : np.ndarray
        Array of matrices the Hermitian conjugate of which needs to be
        determined

    Returns
    -------
    matrix_array_dagger : np.ndarray
        Hermitian conjugate of matrix_array
    """

    if matrix_array.ndim == 2:
        matrix_array_dagger = np.transpose(matrix_array.conj(), axes=(1, 0))

    else:
        matrix_array_dagger = np.transpose(matrix_array.conj(), axes=(0, 2, 1))

    return matrix_array_dagger


def _spectral_factorization(cross_spectrum, num_iterations, term_crit=1e-12):
    r"""
    Determine the spectral matrix factorization of the cross-spectrum
    (denoted by S) of multiple time series:
        S = H \Sigma H^{\dagger}
    Here, \Sigma is the covariance matrix, H the transfer function.
    We follow the algorithm outlined in 'The Factorization of Matricial
    Spectral Densities', Wilson 1972, SiAM J Appl Math.
    The algorithm iteratively calculates approximations for the spectral
    factorization and terminates if either the maximum number of iterations is
    reached or the difference between the cross spectrum and the approximate
    cross spectrum calculated from the covariance matrix and transfer function
    is sufficiently small.

    Parameters
    ----------
    cross_spectrum : np.ndarray
        Cross spectrum to be decomposed in covariance matrix and transfer
        function
    num_iterations : int
        Maximal number of iterations of iterative algorithm
    term_crit : float
        Termination criterion for iteration step in spectral matrix
        factorization
        Default: 1e-12


    Returns
    ------
    cov_matrix : np.ndarray
        Covariance matrix of spectral factorization
    transfer_function : np.ndarray
        Transfer function of spectral factorization
    """
    # spectral_density_function = np.fft.ifft(cross_spectrum, axis=0)
    spectral_density_function = np.copy(cross_spectrum)

    # Resolve dimensions
    num_freqs = np.shape(spectral_density_function)[0]
    num_signals = np.shape(spectral_density_function)[1]

    # Initialization
    identity = np.identity(num_signals)
    factorization = np.zeros(np.shape(spectral_density_function),
                             dtype='complex128')

    # Estimate initial conditions
    try:
        initial_cond = np.linalg.cholesky(cross_spectrum[0].real)
    except np.linalg.LinAlgError:
        raise ValueError('Could not calculate Cholesky decomposition of real'
                         + ' part of zero frequency estimate of cross-spectrum'
                         + '. This might suggest a problem with the input')

    factorization += initial_cond

    converged = False
    # Iteration for calculating spectral factorization
    for i in range(num_iterations):

        factorization_old = np.copy(factorization)

        # Implementation of Eq. 3.1 from "The Factorization of Matricial
        # Spectral Densities", Wilson 1972, SiAM J Appl Math
        X = np.linalg.solve(factorization,
                            spectral_density_function)
        Y = np.linalg.solve(factorization,
                            _dagger(X))
        Y += identity
        Y = _bracket_operator(Y, num_freqs, num_signals)

        factorization = np.matmul(factorization, Y)

        diff = factorization - factorization_old
        error = np.max(np.abs(diff))
        if error < term_crit:
            print(f'Spectral factorization converged after {i} steps')
            converged=True
            break

    if not converged:
        raise Exception("Spectral factorization did not converge after "
                        + f"{num_iterations} steps. Try to increase "
                        + "'num_iterations', or lower the allowed error "
                        + " in the termination criterion, currently "
                        + f"{term_crit}")

    cov_matrix = np.matmul(factorization[0],
                           _dagger(factorization[0]))

    transfer_function = np.matmul(factorization,
                                  np.linalg.inv(factorization[0]))

    return cov_matrix, transfer_function


def pairwise_granger(signals, max_order, information_criterion='aic'):
    r"""
    Determine Granger Causality of two time series

    Parameters
    ----------
    signals : (N, 2) np.ndarray or neo.AnalogSignal
        A matrix with two time series (second dimension) that have N time
        points (first dimension).
    max_order : int
        Maximal order of autoregressive model.
    information_criterion : {'aic', 'bic'}, optional
        A function to compute the information criterion:

        * `bic` for Bayesian information_criterion,
        * `aic` for Akaike information criterion,

        Default: 'aic'

    Returns
    -------
    Causality
        A `namedtuple` with the following attributes:
            directional_causality_x_y : float
                The Granger causality value for X influence onto Y.

            directional_causality_y_x : float
                The Granger causality value for Y influence onto X.

            instantaneous_causality : float
                The remaining channel interdependence not accounted for by
                the directional causalities (e.g. shared input to X and Y).

            total_interdependence : float
                The sum of the former three metrics. It measures the dependence
                of X and Y. If the total interdependence is positive, X and Y
                are not independent.

        Denote covariance matrix of signals
            X by C|X  - a real number
            Y by C|Y - a real number
            (X,Y) by C|XY - a (2 \times 2) matrix

        directional causality X -> Y given by
            log(C|X / C|XY_00)
        directional causality Y -> X given by
            log(C|Y / C|XY_11)
        instantaneous causality of X,Y given by
            log(C|XY_00 / C|XY_11)
        total interdependence of X,Y given by
            log( {C|X \cdot C|Y} / det{C|XY} )

    Raises
    ------
    ValueError
        If the provided signal does not have a shape of Nx2.

        If the determinant of the prediction error covariance matrix is not
        positive.

    Warns
    -----
    UserWarning
        If the log determinant of the prediction error covariance matrix is
        below the tolerance level of 1e-7.

    Notes
    -----
    The formulas used in this implementation follows
    :cite:`granger-Ding06_0608035`. The only difference being that we change
    the equation 47 in the following way:
    -R(k) - A(1)R(k - 1) - ... - A(m)R(k - m) = 0.
    This forumlation allows for the usage of R values without transposition
    (i.e. directly) in equation 48.

    Examples
    --------
    Example 1. Independent variables.

    >>> import numpy as np
    >>> from elephant.causality.granger import pairwise_granger
    >>> pairwise_granger(np.random.uniform(size=(1000, 2)), max_order=2)  # noqa
    Causality(directional_causality_x_y=-0.0, directional_causality_y_x=0.0, instantaneous_causality=0.0, total_interdependence=0.0)

    Example 2. Dependent variables. Y depends on X but not vice versa.

    .. math::
        \begin{array}{ll}
            X_t \sim \mathcal{N}(0, 1) \\
            Y_t = 3.5 \cdot X_{t-1} + \epsilon, \;
                  \epsilon \sim\mathcal{N}(0, 1)
        \end{array}

    In this case, the directional causality is non-zero.

    >>> np.random.seed(31)
    >>> x = np.random.randn(1001)
    >>> y = 3.5 * x[:-1] + np.random.randn(1000)
    >>> signals = np.array([x[1:], y]).T  # N x 2 matrix
    >>> pairwise_granger(signals, max_order=1)  # noqa
    Causality(directional_causality_x_y=2.64, directional_causality_y_x=-0.0, instantaneous_causality=0.0, total_interdependence=2.64)

    """
    if isinstance(signals, AnalogSignal):
        signals = signals.magnitude

    if not (signals.ndim == 2 and signals.shape[1] == 2):
        raise ValueError("The input 'signals' must be of dimensions Nx2.")

    # transpose (N,2) -> (2,N) for mathematical convenience
    signals = signals.T

    # signal_x and signal_y are (1, N) arrays
    signal_x, signal_y = np.expand_dims(signals, axis=1)

    coeffs_x, var_x, p_1 = _optimal_vector_arm(signal_x, 1, max_order,
                                               information_criterion)
    coeffs_y, var_y, p_2 = _optimal_vector_arm(signal_y, 1, max_order,
                                               information_criterion)
    coeffs_xy, cov_xy, p_3 = _optimal_vector_arm(signals, 2, max_order,
                                                 information_criterion)

    sign, log_det_cov = np.linalg.slogdet(cov_xy)
    tolerance = 1e-7

    if sign <= 0:
        raise ValueError(
            "Determinant of covariance matrix must be always positive: "
            "In this case its sign is {}".format(sign))

    if log_det_cov <= tolerance:
        warnings.warn("The value of the log determinant is at or below the "
                      "tolerance level. Proceeding with computation.",
                      UserWarning)

    directional_causality_y_x = np.log(var_x[0]) - np.log(cov_xy[0, 0])
    directional_causality_x_y = np.log(var_y[0]) - np.log(cov_xy[1, 1])

    instantaneous_causality = \
        np.log(cov_xy[0, 0]) + np.log(cov_xy[1, 1]) - log_det_cov
    instantaneous_causality = np.asarray(instantaneous_causality)

    total_interdependence = np.log(var_x[0]) + np.log(var_y[0]) - log_det_cov

    # Round GC according to following scheme:
    #     Note that standard error scales as 1/sqrt(sample_size)
    #     Calculate  significant figures according to standard error
    length = np.size(signal_x)
    asymptotic_std_error = 1/np.sqrt(length)
    est_sig_figures = int((-1)*np.around(np.log10(asymptotic_std_error)))

    directional_causality_x_y_round = np.around(directional_causality_x_y,
                                                est_sig_figures)
    directional_causality_y_x_round = np.around(directional_causality_y_x,
                                                est_sig_figures)
    instantaneous_causality_round = np.around(instantaneous_causality,
                                              est_sig_figures)
    total_interdependence_round = np.around(total_interdependence,
                                            est_sig_figures)

    return Causality(
        directional_causality_x_y=directional_causality_x_y_round.item(),
        directional_causality_y_x=directional_causality_y_x_round.item(),
        instantaneous_causality=instantaneous_causality_round.item(),
        total_interdependence=total_interdependence_round.item())


def conditional_granger(signals, max_order, information_criterion='aic'):
    r"""
    Determine conditional Granger Causality of the second time series on the
    first time series, given the third time series. In other words, for time
    series X_t, Y_t and Z_t, this function tests if Y_t influences X_t via Z_t.

    Parameters
    ----------
    signals : (N, 3) np.ndarray or neo.AnalogSignal
        A matrix with three time series (second dimension) that have N time
        points (first dimension). The time series to be conditioned on is the
        third.
    max_order : int
        Maximal order of autoregressive model.
    information_criterion : {'aic', 'bic'}, optional
        A function to compute the information criterion:

        * `bic` for Bayesian information_criterion,
        * `aic` for Akaike information criterion,

        Default: 'aic'

    Returns
    -------
    conditional_causality_xy_z_round : float
        The value of conditional causality of Y_t on X_t given Z_t. Zero value
        indicates that causality of Y_t on X_t is solely dependent on Z_t.

    Raises
    ------
    ValueError
        If the provided signal does not have a shape of Nx3.

    Notes
    -----
    The formulas used in this implementation follows
    :cite:`granger-Ding06_0608035`. Specifically, the Eq 35.
    """
    if isinstance(signals, AnalogSignal):
        signals = signals.magnitude

    if not (signals.ndim == 2 and signals.shape[1] == 3):
        raise ValueError("The input 'signals' must be of dimensions Nx3.")

    # transpose (N,3) -> (3,N) for mathematical convenience
    signals = signals.T

    # signal_x, signal_y and signal_z are (1, N) arrays
    signal_x, signal_y, signal_z = np.expand_dims(signals, axis=1)

    signals_xz = np.vstack([signal_x, signal_z])

    coeffs_xz, cov_xz, p_1 = _optimal_vector_arm(
        signals_xz, dimension=2, max_order=max_order,
        information_criterion=information_criterion)
    coeffs_xyz, cov_xyz, p_2 = _optimal_vector_arm(
        signals, dimension=3, max_order=max_order,
        information_criterion=information_criterion)

    conditional_causality_xy_z = np.log(cov_xz[0, 0]) - np.log(cov_xyz[0, 0])

    # Round conditional GC according to following scheme:
    #     Note that standard error scales as 1/sqrt(sample_size)
    #     Calculate  significant figures according to standard error
    length = np.size(signal_x)
    asymptotic_std_error = 1/np.sqrt(length)
    est_sig_figures = int((-1)*np.around(np.log10(asymptotic_std_error)))

    conditional_causality_xy_z_round = np.around(conditional_causality_xy_z,
                                                 est_sig_figures)

    return conditional_causality_xy_z_round


def pairwise_spectral_granger(signal_i, signal_j, fs=1, nw=4, num_tapers=None,
                              peak_resolution=None, n_segments=1,
                              len_segment=None, frequency_resolution=None,
                              overlap=0.5, num_iterations=300,
                              term_crit=1e-12):
    r"""Determine spectral Granger Causality of two signals.

    The spectral Granger Causality is obtained through the following steps:

    1. Determine the cross spectrum of the two signals by applying
       :func:`segmented_multitaper_cross_spectrum` to the joint signal. See the
       documentation of this function for the hierarchy of the parameters used
       for the estimation of the cross spectrum.

    2. Calculate the spectral factorization of the cross spectrum decomposing
       the cross spectrum into the covariance matrix and the transfer function.

    3. Calculate the directional and instantaneous spectral Granger Causality
       from the power spectra, the transfer function and the covariance matrix
       (see e.g. Wen et al., 2013, Phil Trans R Soc, eq. 2.10 ff).

    Parameters
    ----------
    signal_i : neo.AnalogSignal or pq.Quantity or np.ndarray
        First time series data of the pair between which spectral Granger
        Causality is computed.
    signal_j : neo.AnalogSignal or pq.Quantity or np.ndarray
        Second time series data of the pair between which spectral Granger
        Causality is computed.
        The shapes and the sampling frequencies of `signal_i` and `signal_j`
        must be identical. When `signal_i` and `signal_j` are not
        `neo.AnalogSignal`, sampling frequency should be specified through the
        keyword argument `fs`. Otherwise, the default value is used
        (`fs` = 1.0).
    fs : float, optional
        Specifies the sampling frequency of the input time series
        Default: 1.0
    nw : float, optional
        Time bandwidth product
        Default: 4.0
    num_tapers : int, optional
        Number of tapers used in 1. to obtain estimate of PSD. By default,
        [2*nw] - 1 is chosen.
        Default: None
    peak_resolution : pq.Quantity float, optional
        Quantity in Hz determining the number of tapers used for analysis.
        Fine peak resolution --> low numerical value --> low number of tapers
        High peak resolution --> high numerical value --> high number of tapers
        When given as a `float`, it is taken as frequency in Hz.
        Default: None.
    n_segments : int, optional
        Number of segments. The length of segments is adjusted so that
        overlapping segments cover the entire stretch of the given data. This
        parameter is ignored if `len_segment` or `frequency_resolution` is
        given.
        Default: 1
    len_segment : int, optional
        Length of segments. This parameter is ignored if `frequency_resolution`
        is given. If None, it will be determined from other parameters.
        Default: None
    frequency_resolution : pq.Quantity or float, optional
        Desired frequency resolution of the obtained spectral Granger Causality
        estimate in terms of the interval between adjacent frequency bins. When
        given as a `float`, it is taken as frequency in Hz.
        If None, it will be determined from other parameters.
        Default: None
    overlap : float, optional
        Overlap between segments represented as a float number between 0 (no
        overlap) and 1 (complete overlap).
        Default: 0.5 (half-overlapped)
    num_iterations : int
        Number of iterations for algorithm to estimate spectral factorization.
        Default: 300
    term_crit : float
        Termination criterion for iteration step in spectral matrix
        factorization
        Default: 1e-12

    Returns
    -------
    freqs : np.ndarray
        Frequencies associated with the spectral Granger Causality estimate.
    Causality
        A `namedtuple` with the following attributes:
            directional_causality_x_y : np.ndarray
                Spectrally resolved Granger causality influence of `signal_i`,
                abbreviated by  X, onto `signal_j`, abbreviated by Y.

            directional_causality_y_x : np.ndarray
                Spectrally resolved Granger causality influence of `signal_j`,
                abbreviated by  Y, onto `signal_i`, abbreviated by Y.

            instantaneous_causality : np.ndarray
                The remaining spectrally resolved channel interdependence not
                accounted for by the directional causalities (e.g. shared input
                to X, i.e. `signal_i`, and Y, i.e. `signal_j`).

            total_interdependence : np.ndarray
                The sum of the former three metrics. It measures the dependence
                of X, i.e. `signal_i` and Y, i.e. `signal_j`. If the total
                interdependence is positive, X and Y are not independent.
    """
    if isinstance(signal_i, neo.core.AnalogSignal) and \
            isinstance(signal_j, neo.core.AnalogSignal):
        signals = signal_i.merge(signal_j)
    elif isinstance(signal_i, np.ndarray) and isinstance(signal_j, np.ndarray):
        signals = np.vstack([signal_i, signal_j])

    # Calculate cross spectrum for signals
    freqs, S = segmented_multitaper_cross_spectrum(
        signals=signals, n_segments=n_segments, len_segment=len_segment,
        frequency_resolution=frequency_resolution, overlap=overlap, fs=fs,
        nw=nw, num_tapers=num_tapers, peak_resolution=peak_resolution,
        return_onesided=False)

    # Remove units attached by the multitaper_cross_spectrum
    if isinstance(S, pq.Quantity):
        S = S.magnitude

    # Transpose cross spectrum due to different conventions used in
    # segemented_multitaper_cross_spectrum and the calculations of spectral
    # Granger causality
    S = np.transpose(S, axes=(1, 0, 2))

    # Move frequencies to first axis - Needed for _spectral_factorization
    S = np.transpose(S, axes=(2, 0, 1))

    # Decompose cross spectrum into covariance and transfer function
    C, H = _spectral_factorization(S, num_iterations=num_iterations)

    # Take positive frequencies
    mask = (freqs >= 0)
    freqs = freqs[mask]

    S = S[mask]
    H = H[mask]

    # Calculate spectral Granger Causality.
    # Formulae follow Wen et al., 2013, Phil Trans R Soc
    H_tilde_xx = H[:, 0, 0] + C[0, 1]/C[0, 0]*H[:, 0, 1]
    H_tilde_yy = H[:, 1, 1] + C[0, 1]/C[1, 1]*H[:, 1, 0]

    directional_causality_y_x = np.log(S[:, 0, 0].real /
                                       (H_tilde_xx
                                        * C[0, 0]
                                        * H_tilde_xx.conj()).real)

    directional_causality_x_y = np.log(S[:, 1, 1].real /
                                       (H_tilde_yy
                                        * C[1, 1]
                                        * H_tilde_yy.conj()).real)

    instantaneous_causality = np.log(
        (H_tilde_xx * C[0, 0] * H_tilde_xx.conj()).real
        * (H_tilde_yy * C[1, 1] * H_tilde_yy.conj()).real)
    instantaneous_causality -= np.linalg.slogdet(S)[1]

    total_interdependence = (directional_causality_x_y
                             + directional_causality_y_x
                             + instantaneous_causality)

    spectral_causality = Causality(
        directional_causality_x_y=directional_causality_x_y,
        directional_causality_y_x=directional_causality_y_x,
        instantaneous_causality=instantaneous_causality,
        total_interdependence=total_interdependence)

    return freqs, spectral_causality



# -*- coding: utf-8 -*-
"""
Unit tests for the GPFA analysis.

:copyright: Copyright 2014-2024 by the Elephant team, see AUTHORS.txt.
:license: Modified BSD, see LICENSE.txt for details.
"""

import unittest

import neo
import numpy as np
import quantities as pq
from numpy.testing import assert_array_equal, assert_array_almost_equal

from elephant.spike_train_generation import StationaryPoissonProcess
from elephant.trials import TrialsFromLists
try:
    import sklearn
    HAVE_SKLEARN = True
except ModuleNotFoundError:
    HAVE_SKLEARN = False

if HAVE_SKLEARN:
    from elephant.gpfa import gpfa_util
    from elephant.gpfa import GPFA
    from sklearn.model_selection import cross_val_score


@unittest.skipUnless(HAVE_SKLEARN, 'requires sklearn')
class GPFATestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        def gen_gamma_spike_train(k, theta, t_max):
            x = []
            for i in range(int(3 * t_max / (k * theta))):
                x.append(np.random.gamma(k, theta))
            s = np.cumsum(x)
            return s[s < t_max]

        def gen_test_data(rates, durs, shapes=(1, 1, 1, 1)):
            s = gen_gamma_spike_train(shapes[0], 1. / rates[0], durs[0])
            for i in range(1, 4):
                s_i = gen_gamma_spike_train(shapes[i], 1. / rates[i], durs[i])
                s = np.concatenate([s, s_i + np.sum(durs[:i])])
            return s

        cls.n_iters = 10
        cls.bin_size = 20 * pq.ms

        # generate data1
        rates_a = (2, 10, 2, 2)
        rates_b = (2, 2, 10, 2)
        durs = (2.5, 2.5, 2.5, 2.5)
        np.random.seed(0)
        n_trials = 100
        cls.data0 = []
        for trial in range(n_trials):
            n1 = neo.SpikeTrain(gen_test_data(rates_a, durs), units=1 * pq.s,
                                t_start=0 * pq.s, t_stop=10 * pq.s)
            n2 = neo.SpikeTrain(gen_test_data(rates_a, durs), units=1 * pq.s,
                                t_start=0 * pq.s, t_stop=10 * pq.s)
            n3 = neo.SpikeTrain(gen_test_data(rates_b, durs), units=1 * pq.s,
                                t_start=0 * pq.s, t_stop=10 * pq.s)
            n4 = neo.SpikeTrain(gen_test_data(rates_b, durs), units=1 * pq.s,
                                t_start=0 * pq.s, t_stop=10 * pq.s)
            n5 = neo.SpikeTrain(gen_test_data(rates_a, durs), units=1 * pq.s,
                                t_start=0 * pq.s, t_stop=10 * pq.s)
            n6 = neo.SpikeTrain(gen_test_data(rates_a, durs), units=1 * pq.s,
                                t_start=0 * pq.s, t_stop=10 * pq.s)
            n7 = neo.SpikeTrain(gen_test_data(rates_b, durs), units=1 * pq.s,
                                t_start=0 * pq.s, t_stop=10 * pq.s)
            n8 = neo.SpikeTrain(gen_test_data(rates_b, durs), units=1 * pq.s,
                                t_start=0 * pq.s, t_stop=10 * pq.s)
            cls.data0.append([n1, n2, n3, n4, n5, n6, n7, n8])
        cls.x_dim = 4

        cls.data1 = cls.data0[:20]

        # generate data2
        np.random.seed(27)
        cls.data2 = []
        n_trials = 10
        n_channels = 20
        for trial in range(n_trials):
            rates = np.random.randint(low=1, high=100, size=n_channels)
            spike_times = [StationaryPoissonProcess(rate=rate * pq.Hz,
                           t_stop=1000.0 * pq.ms).generate_spiketrain()
                           for rate in rates]
            cls.data2.append(spike_times)

    def test_data1(self):
        gpfa = GPFA(x_dim=self.x_dim, em_max_iters=self.n_iters)
        gpfa.fit(self.data1)
        latent_variable_orth = gpfa.transform(self.data1)
        self.assertAlmostEqual(gpfa.transform_info['log_likelihood'],
                               -8172.004695554373, places=5)
        # Since data1 is inherently 2 dimensional, only the first two
        # dimensions of latent_variable_orth should have finite power.
        for i in [0, 1]:
            self.assertNotEqual(latent_variable_orth[0][i].mean(), 0)
            self.assertNotEqual(latent_variable_orth[0][i].var(), 0)
        for i in [2, 3]:
            self.assertAlmostEqual(latent_variable_orth[0][i].mean(), 0,
                                   places=2)
            self.assertAlmostEqual(latent_variable_orth[0][i].var(), 0,
                                   places=2)

    def test_transform_testing_data(self):
        # check if the num. of neurons in the test data matches the
        # num. of neurons in the training data
        gpfa1 = GPFA(x_dim=self.x_dim, em_max_iters=self.n_iters)
        gpfa1.fit(self.data1)
        with self.assertRaises(ValueError):
            gpfa1.transform(self.data2)

    def test_cross_validation(self):
        # If GPFA.__init__ is decorated, sklearn signature function parsing
        # magic throws the error
        # __init__() got an unexpected keyword argument 'args'
        lls = []
        for x_dim in range(1, self.x_dim + 1):
            gpfa = GPFA(x_dim=x_dim, em_max_iters=self.n_iters)
            lls.append(np.mean(cross_val_score(gpfa, self.data1, cv=5)))
        self.assertTrue(np.argmax(lls) == 1)

    def test_invalid_input_data(self):
        invalid_data = [(0, [0, 1, 2])]
        invalid_bin_size = 10
        invalid_tau_init = 100
        with self.assertRaises(ValueError):
            _ = GPFA(bin_size=invalid_bin_size)
        with self.assertRaises(ValueError):
            _ = GPFA(tau_init=invalid_tau_init)
        gpfa = GPFA()
        with self.assertRaises(ValueError):
            gpfa.fit(spiketrains=invalid_data)
        with self.assertRaises(ValueError):
            gpfa.fit(spiketrains=[])

    def test_data2(self):
        gpfa = GPFA(bin_size=self.bin_size, x_dim=8, em_max_iters=self.n_iters)
        gpfa.fit(self.data2)
        n_trials = len(self.data2)
        returned_data = gpfa.valid_data_names
        seqs = gpfa.transform(self.data2, returned_data=returned_data)
        for key, data in seqs.items():
            self.assertEqual(len(data), n_trials,
                             msg="Failed ndarray field {0}".format(key))
        t_start = self.data2[0][0].t_stop
        t_stop = self.data2[0][0].t_start
        n_bins = int(((t_start - t_stop) / self.bin_size).magnitude)
        assert_array_equal(gpfa.transform_info['num_bins'],
                           [n_bins, ] * n_trials)

    def test_returned_data(self):
        gpfa = GPFA(bin_size=self.bin_size, x_dim=8, em_max_iters=self.n_iters)
        gpfa.fit(self.data2)
        seqs = gpfa.transform(self.data2)
        self.assertTrue(isinstance(seqs, np.ndarray))

        returned_data = gpfa.valid_data_names
        seqs = gpfa.transform(self.data2, returned_data=returned_data)
        self.assertTrue(len(returned_data) == len(seqs))
        self.assertTrue(isinstance(seqs, dict))
        with self.assertRaises(ValueError):
            seqs = gpfa.transform(self.data2, returned_data=['invalid_name'])

    def test_fit_transform(self):
        gpfa1 = GPFA(bin_size=self.bin_size, x_dim=self.x_dim,
                     em_max_iters=self.n_iters)
        gpfa1.fit(self.data1)
        latent_variable_orth1 = gpfa1.transform(self.data1)
        latent_variable_orth2 = GPFA(
            bin_size=self.bin_size, x_dim=self.x_dim,
            em_max_iters=self.n_iters).fit_transform(self.data1)
        for i in range(len(self.data1)):
            for j in range(self.x_dim):
                assert_array_almost_equal(latent_variable_orth1[i][j],
                                          latent_variable_orth2[i][j])

    def test_get_seq_sqrt(self):
        data = [self.data2[0]]
        seqs = gpfa_util.get_seqs(data, bin_size=self.bin_size)
        seqs_not_sqrt = gpfa_util.get_seqs(data, bin_size=self.bin_size,
                                           use_sqrt=False)
        self.assertEqual(seqs['T'], seqs_not_sqrt['T'])
        self.assertEqual(seqs['y'].shape, seqs_not_sqrt['y'].shape)

    def test_cut_trials_inf(self):
        same_data = gpfa_util.cut_trials(self.data2, seg_length=np.Inf)
        assert same_data is self.data2

    def test_cut_trials_zero_length(self):
        seqs = gpfa_util.get_seqs(self.data2, bin_size=self.bin_size)
        with self.assertRaises(ValueError):
            gpfa_util.cut_trials(seqs, seg_length=0)

    def test_cut_trials_same_length(self):
        data = [self.data2[0]]
        seqs = gpfa_util.get_seqs(data, bin_size=self.bin_size)
        seg_length = seqs[0]['T']
        seqs_cut = gpfa_util.cut_trials(seqs, seg_length=seg_length)
        assert_array_almost_equal(seqs[0]['y'], seqs_cut[0]['y'])

    def test_cut_trials_larger_length(self):
        data = [self.data2[0]]
        seqs = gpfa_util.get_seqs(data, bin_size=self.bin_size)
        seg_length = seqs[0]['T'] + 1
        with self.assertWarns(UserWarning):
            gpfa_util.cut_trials(seqs, seg_length=seg_length)

    def test_logdet(self):
        np.random.seed(27)
        # generate a positive definite matrix
        matrix = np.random.randn(20, 20)
        matrix = matrix.dot(matrix.T)
        logdet_fast = gpfa_util.logdet(matrix)
        logdet_ground_truth = np.log(np.linalg.det(matrix))
        assert_array_almost_equal(logdet_fast, logdet_ground_truth)

    def test_trial_object_gpfa_fit(self):
        gpfa_trial_object = GPFA(bin_size=self.bin_size, x_dim=self.x_dim,
                                 em_max_iters=self.n_iters)
        gpfa_list_of_lists = GPFA(bin_size=self.bin_size, x_dim=self.x_dim,
                                  em_max_iters=self.n_iters)

        trials = TrialsFromLists(self.data1)
        gpfa_trial_object.fit(trials)
        gpfa_list_of_lists.fit(self.data1)

        assert_array_almost_equal(gpfa_trial_object.params_estimated['gamma'],
                                  gpfa_list_of_lists.params_estimated['gamma'])

    def test_trial_object_gpfa_transform(self):
        gpfa_trial_object = GPFA(bin_size=self.bin_size, x_dim=self.x_dim,
                                 em_max_iters=self.n_iters)
        gpfa_list_of_lists = GPFA(bin_size=self.bin_size, x_dim=self.x_dim,
                                  em_max_iters=self.n_iters)

        trials = TrialsFromLists(self.data1)
        gpfa_trial_object.fit(trials)
        gpfa_trial_object.transform(trials)
        gpfa_list_of_lists.fit(self.data1)
        gpfa_list_of_lists.transform(self.data1)

        assert_array_almost_equal(gpfa_trial_object.transform_info['Corth'],
                                  gpfa_list_of_lists.transform_info['Corth'])

    def test_trial_object_zero_trials(self):
        gpfa_trial_object = GPFA(bin_size=self.bin_size, x_dim=self.x_dim,
                                 em_max_iters=self.n_iters)

        trials = TrialsFromLists([])
        with self.assertRaises(ValueError):
            gpfa_trial_object.fit(trials)


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
Unit tests for the spade module.

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""
import unittest

import neo
from neo.core.spiketrainlist import SpikeTrainList
import numpy as np
import quantities as pq
from numpy.testing import assert_array_equal, assert_equal

import elephant.conversion as conv
import elephant.spade as spade
from elephant.spike_train_generation import StationaryPoissonProcess, \
    compound_poisson_process

try:
    import statsmodels

    HAVE_STATSMODELS = True
except ImportError:
    HAVE_STATSMODELS = False

HAVE_FIM = spade.HAVE_FIM


class SpadeTestCase(unittest.TestCase):
    def setUp(self):
        np.random.seed(0)
        # Spade parameters
        self.bin_size = 1 * pq.ms
        self.winlen = 10
        self.n_subset = 10
        self.n_surr = 10
        self.alpha = 0.05
        self.stability_thresh = [0.1, 0.1]
        self.psr_param = [0, 0, 0]
        self.min_occ = 4
        self.min_spikes = 4
        self.max_occ = 4
        self.max_spikes = 4
        self.min_neu = 4
        # Test data parameters
        # CPP parameters
        self.n_neu = 100
        self.amplitude = [0] * self.n_neu + [1]
        self.cpp = compound_poisson_process(
            rate=3 * pq.Hz,
            amplitude_distribution=self.amplitude,
            t_stop=5 * pq.s)
        # Number of patterns' occurrences
        self.n_occ1 = 10
        self.n_occ2 = 12
        self.n_occ3 = 15
        # Patterns lags
        self.lags1 = [2]
        self.lags2 = [1, 3]
        self.lags3 = [1, 2, 4, 5, 7]
        # Length of the spiketrain
        self.t_stop = 3000
        # Patterns times
        self.patt1_times = neo.SpikeTrain(
            np.arange(
                0, 1000, 1000 // self.n_occ1) *
            pq.ms, t_stop=self.t_stop * pq.ms)
        self.patt2_times = neo.SpikeTrain(
            np.arange(
                1000, 2000, 1000 // self.n_occ2)[:-1] *
            pq.ms, t_stop=self.t_stop * pq.ms)
        self.patt3_times = neo.SpikeTrain(
            np.arange(
                2000, 3000, 1000 // self.n_occ3)[:-1] *
            pq.ms, t_stop=self.t_stop * pq.ms)
        # Patterns
        self.patt1 = [self.patt1_times] + [neo.SpikeTrain(
            self.patt1_times.view(pq.Quantity) + lag * pq.ms,
            t_stop=self.t_stop * pq.ms) for lag in self.lags1]
        self.patt2 = [self.patt2_times] + [neo.SpikeTrain(
            self.patt2_times.view(pq.Quantity) + lag * pq.ms,
            t_stop=self.t_stop * pq.ms) for lag in self.lags2]
        self.patt3 = [self.patt3_times] + [neo.SpikeTrain(
            self.patt3_times.view(pq.Quantity) + lag * pq.ms,
            t_stop=self.t_stop * pq.ms) for lag in self.lags3]
        # Data
        self.msip = self.patt1 + self.patt2 + self.patt3
        # Expected results
        self.n_spk1 = len(self.lags1) + 1
        self.n_spk2 = len(self.lags2) + 1
        self.n_spk3 = len(self.lags3) + 1
        self.elements1 = list(range(self.n_spk1))
        self.elements2 = list(range(self.n_spk2))
        self.elements3 = list(range(self.n_spk3))
        self.elements_msip = [
            self.elements1,
            list(
                range(
                    self.n_spk1,
                    self.n_spk1 +
                    self.n_spk2)),
            list(
                range(
                    self.n_spk1 +
                    self.n_spk2,
                    self.n_spk1 +
                    self.n_spk2 +
                    self.n_spk3))]
        self.occ1 = np.unique(conv.BinnedSpikeTrain(
            self.patt1_times, self.bin_size).spike_indices[0])
        self.occ2 = np.unique(conv.BinnedSpikeTrain(
            self.patt2_times, self.bin_size).spike_indices[0])
        self.occ3 = np.unique(conv.BinnedSpikeTrain(
            self.patt3_times, self.bin_size).spike_indices[0])
        self.occ_msip = [
            list(self.occ1), list(self.occ2), list(self.occ3)]
        self.lags_msip = [self.lags1, self.lags2, self.lags3]
        self.patt_psr = self.patt3 + [self.patt3[-1][:3]]

    # Testing cpp
    @unittest.skipUnless(HAVE_FIM, "Time consuming with pythonic FIM")
    def test_spade_cpp(self):
        output_cpp = spade.spade(self.cpp, self.bin_size, 1,
                                 approx_stab_pars=dict(
                                     n_subsets=self.n_subset,
                                     stability_thresh=self.stability_thresh),
                                 n_surr=self.n_surr, alpha=self.alpha,
                                 psr_param=self.psr_param,
                                 stat_corr='no',
                                 output_format='patterns')['patterns']
        elements_cpp = []
        lags_cpp = []
        # collecting spade output
        for out in output_cpp:
            elements_cpp.append(sorted(out['neurons']))
            lags_cpp.append(list(out['lags'].magnitude))
        # check neurons in the patterns
        assert_array_equal(elements_cpp, [range(self.n_neu)])
        # check the lags
        assert_array_equal(lags_cpp, [np.array([0] * (self.n_neu - 1))])

    # Testing spectrum cpp
    def test_spade_spectrum_cpp(self):
        # Computing Spectrum
        spectrum_cpp = spade.concepts_mining(self.cpp, self.bin_size,
                                             1, report='#')[0]
        # Check spectrum
        assert_array_equal(
            spectrum_cpp,
            [(len(self.cpp),
              np.sum(conv.BinnedSpikeTrain(
                  self.cpp[0], self.bin_size).to_bool_array()), 1)])

    # Testing with multiple patterns input
    def test_spade_msip(self):
        output_msip = spade.spade(self.msip, self.bin_size, self.winlen,
                                  approx_stab_pars=dict(
                                      n_subsets=self.n_subset,
                                      stability_thresh=self.stability_thresh),
                                  n_surr=self.n_surr, alpha=self.alpha,
                                  psr_param=self.psr_param,
                                  stat_corr='no',
                                  output_format='patterns')['patterns']
        elements_msip = []
        occ_msip = []
        lags_msip = []
        # collecting spade output
        for out in output_msip:
            elements_msip.append(out['neurons'])
            occ_msip.append(list(out['times'].magnitude))
            lags_msip.append(list(out['lags'].magnitude))
        elements_msip = sorted(elements_msip, key=len)
        occ_msip = sorted(occ_msip, key=len)
        lags_msip = sorted(lags_msip, key=len)
        # check neurons in the patterns
        assert_equal(elements_msip, self.elements_msip)
        # check the occurrences time of the patters
        assert_equal(occ_msip, self.occ_msip)
        # check the lags
        assert_equal(lags_msip, self.lags_msip)

    # Testing with multiple patterns input
    def test_spade_msip_spiketrainlist(self):
        output_msip = spade.spade(
            SpikeTrainList(self.msip), self.bin_size, self.winlen,
            approx_stab_pars=dict(n_subsets=self.n_subset,
                                  stability_thresh=self.stability_thresh),
            n_surr=self.n_surr,
            alpha=self.alpha,
            psr_param=self.psr_param,
            stat_corr='no',
            output_format='patterns')['patterns']
        elements_msip = []
        occ_msip = []
        lags_msip = []
        # collecting spade output
        for out in output_msip:
            elements_msip.append(out['neurons'])
            occ_msip.append(list(out['times'].magnitude))
            lags_msip.append(list(out['lags'].magnitude))
        elements_msip = sorted(elements_msip, key=len)
        occ_msip = sorted(occ_msip, key=len)
        lags_msip = sorted(lags_msip, key=len)
        # check neurons in the patterns
        assert_equal(elements_msip, self.elements_msip)
        # check the occurrences time of the patters
        assert_equal(occ_msip, self.occ_msip)
        # check the lags
        assert_equal(lags_msip, self.lags_msip)

    def test_parameters(self):
        """
        Test under different configuration of parameters than the default one
        """
        # test min_spikes parameter
        with self.assertWarns(UserWarning):
            # n_surr=0 and alpha=0.05 spawns expected UserWarning
            output_msip_min_spikes = spade.spade(
                self.msip,
                self.bin_size,
                self.winlen,
                min_spikes=self.min_spikes,
                approx_stab_pars=dict(n_subsets=self.n_subset),
                n_surr=0,
                alpha=self.alpha,
                psr_param=self.psr_param,
                stat_corr='no',
                output_format='patterns')['patterns']
        # collecting spade output
        elements_msip_min_spikes = []
        for out in output_msip_min_spikes:
            elements_msip_min_spikes.append(out['neurons'])
        elements_msip_min_spikes = sorted(
            elements_msip_min_spikes, key=len)
        lags_msip_min_spikes = []
        for out in output_msip_min_spikes:
            lags_msip_min_spikes.append(list(out['lags'].magnitude))
            pvalue = out['pvalue']
        lags_msip_min_spikes = sorted(
            lags_msip_min_spikes, key=len)
        # check the lags
        assert_array_equal(lags_msip_min_spikes, [
            lag for lag in self.lags_msip if len(lag) + 1 >= self.min_spikes])
        # check the neurons in the patterns
        assert_array_equal(elements_msip_min_spikes, [
            el for el in self.elements_msip if len(el) >= self.min_neu and len(
                el) >= self.min_spikes])
        # check that the p-values assigned are equal to -1 (n_surr=0)
        assert_array_equal(-1, pvalue)

        # test min_occ parameter
        output_msip_min_occ = spade.spade(
            self.msip,
            self.bin_size,
            self.winlen,
            min_occ=self.min_occ,
            approx_stab_pars=dict(
                n_subsets=self.n_subset),
            n_surr=self.n_surr,
            alpha=self.alpha,
            psr_param=self.psr_param,
            stat_corr='no',
            output_format='patterns')['patterns']
        # collect spade output
        occ_msip_min_occ = []
        for out in output_msip_min_occ:
            occ_msip_min_occ.append(list(out['times'].magnitude))
        occ_msip_min_occ = sorted(occ_msip_min_occ, key=len)
        # test occurrences time
        assert_equal(occ_msip_min_occ, [
            occ for occ in self.occ_msip if len(occ) >= self.min_occ])

        # test max_spikes parameter
        output_msip_max_spikes = spade.spade(
            self.msip,
            self.bin_size,
            self.winlen,
            max_spikes=self.max_spikes,
            approx_stab_pars=dict(
                n_subsets=self.n_subset),
            n_surr=self.n_surr,
            alpha=self.alpha,
            psr_param=self.psr_param,
            stat_corr='no',
            output_format='patterns')['patterns']
        # collecting spade output
        elements_msip_max_spikes = []
        for out in output_msip_max_spikes:
            elements_msip_max_spikes.append(out['neurons'])
        lags_msip_max_spikes = []
        for out in output_msip_max_spikes:
            lags_msip_max_spikes.append(list(out['lags'].magnitude))
        lags_msip_max_spikes = sorted(
            lags_msip_max_spikes, key=len)
        # check the lags
        assert_array_equal(
            [len(lags) < self.max_spikes
             for lags in lags_msip_max_spikes],
            [True] * len(lags_msip_max_spikes))

        # TODO: does not work with new FIM module
        # test max_occ parameter
        # output_msip_max_occ = spade.spade(
        #     self.msip,
        #     self.bin_size,
        #     self.winlen,
        #     max_occ=self.max_occ,
        #     approx_stab_pars=dict(
        #         n_subsets=self.n_subset),
        #     n_surr=self.n_surr,
        #     alpha=self.alpha,
        #     psr_param=self.psr_param,
        #     stat_corr='no',
        #     output_format='patterns')['patterns']
        # # collect spade output
        # occ_msip_max_occ = []
        # for out in output_msip_max_occ:
        #     occ_msip_max_occ.append(list(out['times'].magnitude))
        # occ_msip_max_occ = sorted(occ_msip_max_occ, key=len)
        # # test occurrences time
        # assert_array_equal(occ_msip_max_occ, [
        #     occ for occ in self.occ_msip if len(occ) <= self.max_occ])

    # test to compare the python and the C implementation of FIM
    # skip this test if C code not available
    @unittest.skipIf(not HAVE_FIM, 'Requires fim.so')
    def test_fpgrowth_fca(self):
        print("fim.so is found.")
        binary_matrix = conv.BinnedSpikeTrain(
            self.patt1, self.bin_size).to_sparse_bool_array().tocoo()
        context, transactions, rel_matrix = spade._build_context(
            binary_matrix, self.winlen)
        # mining the data with python fast_fca
        mining_results_fpg = spade._fpgrowth(
            transactions,
            rel_matrix=rel_matrix)
        print(
            '################################################################')
        print('mining results fpg', mining_results_fpg)
        # mining the data with C fim
        mining_results_ffca = spade._fast_fca(context)

        # testing that the outputs are identical
        assert_array_equal(sorted(mining_results_ffca[0][0]), sorted(
            mining_results_fpg[0][0]))
        assert_array_equal(sorted(mining_results_ffca[0][1]), sorted(
            mining_results_fpg[0][1]))

    def test_spade_output_for_specific_pattern(self):

        np.random.seed(0)

        n_spiketrains = 3
        spiketrains = StationaryPoissonProcess(
            rate=20 * pq.Hz, t_stop=5 * pq.s).generate_n_spiketrains(
            n_spiketrains=n_spiketrains)
        bin_size = 5 * pq.ms

        spade_output = spade.spade(
            spiketrains,
            bin_size=bin_size,
            winlen=2,
            min_spikes=3,
            max_spikes=3,
            min_neu=n_spiketrains,
            spectrum='3d#')

        patterns = spade_output['patterns']
        self.assertEqual (len(patterns), 1)
        pattern = patterns[0]
        self.assertEqual(len(np.unique(pattern['neurons'])), n_spiketrains)
        assert_array_equal(sorted(pattern['itemset']), [1, 2, 5])
        assert_array_equal(sorted(pattern['windows_ids']), [36, 134, 856])
        assert_array_equal(pattern['signature'], [3, 3, 1])

    # Tests 3d spectrum
    # Testing with multiple patterns input
    def test_spade_msip_3d(self):
        output_msip = spade.spade(self.msip, self.bin_size, self.winlen,
                                  approx_stab_pars=dict(
                                      n_subsets=self.n_subset,
                                      stability_thresh=self.stability_thresh),
                                  n_surr=self.n_surr, spectrum='3d#',
                                  alpha=self.alpha, psr_param=self.psr_param,
                                  stat_corr='no',
                                  output_format='patterns')['patterns']
        elements_msip = []
        occ_msip = []
        lags_msip = []
        # collecting spade output
        for out in output_msip:
            elements_msip.append(out['neurons'])
            occ_msip.append(list(out['times'].magnitude))
            lags_msip.append(list(out['lags'].magnitude))
        elements_msip = sorted(elements_msip, key=len)
        occ_msip = sorted(occ_msip, key=len)
        lags_msip = sorted(lags_msip, key=len)
        # check neurons in the patterns
        assert_equal(elements_msip, self.elements_msip)
        # check the occurrences time of the patters
        assert_equal(occ_msip, self.occ_msip)
        # check the lags
        assert_equal(lags_msip, self.lags_msip)

    # test under different configuration of parameters than the default one
    def test_parameters_3d(self):
        # test min_spikes parameter
        output_msip_min_spikes = spade.spade(
            self.msip,
            self.bin_size,
            self.winlen,
            min_spikes=self.min_spikes,
            approx_stab_pars=dict(
                n_subsets=self.n_subset),
            n_surr=self.n_surr,
            spectrum='3d#',
            alpha=self.alpha,
            psr_param=self.psr_param,
            stat_corr='no',
            output_format='patterns')['patterns']
        # collecting spade output
        elements_msip_min_spikes = []
        for out in output_msip_min_spikes:
            elements_msip_min_spikes.append(out['neurons'])
        elements_msip_min_spikes = sorted(
            elements_msip_min_spikes, key=len)
        lags_msip_min_spikes = []
        for out in output_msip_min_spikes:
            lags_msip_min_spikes.append(list(out['lags'].magnitude))
        lags_msip_min_spikes = sorted(
            lags_msip_min_spikes, key=len)
        # check the lags
        assert_array_equal(lags_msip_min_spikes, [
            lag for lag in self.lags_msip if len(lag) + 1 >= self.min_spikes])
        # check the neurons in the patterns
        assert_array_equal(elements_msip_min_spikes, [
            el for el in self.elements_msip if len(el) >= self.min_neu and len(
                el) >= self.min_spikes])

        # test min_occ parameter
        output_msip_min_occ = spade.spade(
            self.msip,
            self.bin_size,
            self.winlen,
            min_occ=self.min_occ,
            approx_stab_pars=dict(
                n_subsets=self.n_subset),
            n_surr=self.n_surr,
            spectrum='3d#',
            alpha=self.alpha,
            psr_param=self.psr_param,
            stat_corr='no',
            output_format='patterns')['patterns']
        # collect spade output
        occ_msip_min_occ = []
        for out in output_msip_min_occ:
            occ_msip_min_occ.append(list(out['times'].magnitude))
        occ_msip_min_occ = sorted(occ_msip_min_occ, key=len)
        # test occurrences time
        assert_equal(occ_msip_min_occ, [
            occ for occ in self.occ_msip if len(occ) >= self.min_occ])

    # Test computation spectrum
    def test_spectrum(self):
        # test 2d spectrum
        spectrum = spade.concepts_mining(self.patt1, self.bin_size,
                                         self.winlen, report='#')[0]
        # test 3d spectrum
        assert_array_equal(spectrum, [[len(self.lags1) + 1, self.n_occ1, 1]])
        spectrum_3d = spade.concepts_mining(self.patt1, self.bin_size,
                                            self.winlen, report='3d#')[0]
        assert_array_equal(spectrum_3d, [
            [len(self.lags1) + 1, self.n_occ1, max(self.lags1), 1]])

    def test_spade_raise_error(self):
        # Test list not using neo.Spiketrain
        self.assertRaises(TypeError, spade.spade, [
            [1, 2, 3], [3, 4, 5]], 1 * pq.ms, 4, stat_corr='no')
        self.assertRaises(TypeError, spade.concepts_mining, [
            [1, 2, 3], [3, 4, 5]], 1 * pq.ms, 4)
        # Test neo.Spiketrain with different t_stop
        self.assertRaises(
            ValueError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=5 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            1 * pq.ms, 4, stat_corr='no')
        # Test bin_size not pq.Quantity
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1., winlen=4, stat_corr='no')
        # Test winlen not int
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4.1, stat_corr='no')
        # Test min_spikes not int
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, min_spikes=3.4, stat_corr='no')
        # Test min_occ not int
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, min_occ=3.4, stat_corr='no')
        # Test max_spikes not int
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, max_spikes=3.4, stat_corr='no')
        # Test max_occ not int
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, max_occ=3.4, stat_corr='no')
        # Test min_neu not int
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, min_neu=3.4, stat_corr='no')
        # Test wrong stability params
        self.assertRaises(
            ValueError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, approx_stab_pars={'wrong key': 0},
            stat_corr='no')
        # Test n_surr not int
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, n_surr=3.4, stat_corr='no')
        # Test dither not pq.Quantity
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, n_surr=100, alpha=0.05,
            dither=15., stat_corr='no')
        # Test wrong alpha
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, n_surr=100, alpha='5 %',
            dither=15. * pq.ms, stat_corr='no')
        # Test wrong statistical correction
        self.assertRaises(
            ValueError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, n_surr=100, alpha=0.05,
            dither=15. * pq.ms, stat_corr='wrong correction')
        # Test wrong psr_params
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, n_surr=100, alpha=0.05,
            dither=15. * pq.ms, stat_corr='no', psr_param=(2.5, 3.4, 2.1))
        # Test wrong psr_params
        self.assertRaises(
            TypeError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, n_surr=100, alpha=0.05,
            dither=15. * pq.ms, stat_corr='no', psr_param=3.1)
        # Test output format
        self.assertRaises(
            ValueError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            bin_size=1. * pq.ms, winlen=4, n_surr=100, alpha=0.05,
            dither=15. * pq.ms, stat_corr='no', output_format='wrong_output')
        # Test wrong spectrum parameter
        self.assertRaises(
            ValueError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            1 * pq.ms, 4, n_surr=1, stat_corr='no',
            spectrum='invalid_key')
        self.assertRaises(
            ValueError, spade.concepts_mining,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            1 * pq.ms, 4, report='invalid_key')
        self.assertRaises(
            ValueError, spade.pvalue_spectrum,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=6 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=6 * pq.s)],
            1 * pq.ms, 4, dither=10 * pq.ms, n_surr=1,
            spectrum='invalid_key')
        # Test negative minimum number of spikes
        self.assertRaises(
            ValueError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=5 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=5 * pq.s)],
            1 * pq.ms, 4, min_neu=-3, stat_corr='no')
        # Test wrong dither method
        self.assertRaises(
            ValueError, spade.spade,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=5 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=5 * pq.s)],
            1 * pq.ms, 4, surr_method='invalid_key', stat_corr='no')
        # Test negative number of surrogates
        self.assertRaises(
            ValueError, spade.pvalue_spectrum,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=5 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=5 * pq.s)],
            1 * pq.ms, 4, dither=10 * pq.ms, n_surr=100,
            surr_method='invalid_key')
        # Test negative number of surrogates
        self.assertRaises(
            ValueError, spade.pvalue_spectrum,
            [neo.SpikeTrain([1, 2, 3] * pq.s, t_stop=5 * pq.s),
             neo.SpikeTrain([3, 4, 5] * pq.s, t_stop=5 * pq.s)],
            1 * pq.ms, 4, 3 * pq.ms, n_surr=-3)
        # Test wrong correction parameter
        self.assertRaises(ValueError, spade.test_signature_significance,
                          pv_spec=((2, 3, 0.2), (2, 4, 0.1)),
                          concepts=([[(2, 3), (1, 2, 3)]]),
                          alpha=0.01,
                          winlen=1,
                          corr='invalid_key')
        # Test negative number of subset for stability
        self.assertRaises(ValueError, spade.approximate_stability, (),
                          np.ndarray([]), n_subsets=-3)

    def test_pattern_set_reduction(self):
        winlen = 6
        # intent(concept1) is a superset of intent(concept2)
        # extent(concept1) is a subset of extent(concept2)
        # intent(concept2) is a subset of intent(concept3)
        #     when taking into account the shift due to the window positions
        # intent(concept1) has a non-empty intersection with intent(concept3)
        #     when taking into account the shift due to the window positions
        # intent(concept4) is disjoint from all others
        concept1 = ((12, 19, 26), (2, 10, 18))
        concept2 = ((12, 19), (2, 10, 18, 26))
        concept3 = ((0, 7, 14, 21), (0, 8))
        concept4 = ((1, 6), (0, 8))

        # reject concept2 using min_occ
        # make sure to keep concept1 by setting k_superset_filtering = 1
        concepts = spade.pattern_set_reduction([concept1, concept2],
                                               ns_signatures=[],
                                               winlen=winlen, spectrum='#',
                                               h_subset_filtering=0, min_occ=2,
                                               k_superset_filtering=1)
        self.assertEqual(concepts, [concept1])

        # keep concept2 by increasing h_subset_filtering
        concepts = spade.pattern_set_reduction([concept1, concept2],
                                               ns_signatures=[],
                                               winlen=winlen, spectrum='#',
                                               h_subset_filtering=2, min_occ=2,
                                               k_superset_filtering=1)
        self.assertEqual(concepts, [concept1, concept2])

        # reject concept1 using min_spikes
        concepts = spade.pattern_set_reduction([concept1, concept2],
                                               ns_signatures=[],
                                               winlen=winlen, spectrum='#',
                                               h_subset_filtering=2,
                                               min_spikes=2,
                                               k_superset_filtering=0)
        self.assertEqual(concepts, [concept2])

        # reject concept2 using ns_signatures
        concepts = spade.pattern_set_reduction([concept1, concept2],
                                               ns_signatures=[(2, 2)],
                                               winlen=winlen, spectrum='#',
                                               h_subset_filtering=1, min_occ=2,
                                               k_superset_filtering=1)
        self.assertEqual(concepts, [concept1])

        # reject concept1 using ns_signatures
        # make sure to keep concept2 by increasing h_subset_filtering
        concepts = spade.pattern_set_reduction([concept1, concept2],
                                               ns_signatures=[(2, 3)],
                                               winlen=winlen, spectrum='#',
                                               h_subset_filtering=3,
                                               min_spikes=2,
                                               min_occ=2,
                                               k_superset_filtering=1)
        self.assertEqual(concepts, [concept2])

        # reject concept2 using the covered spikes criterion
        concepts = spade.pattern_set_reduction([concept1, concept2],
                                               ns_signatures=[(2, 2)],
                                               winlen=winlen, spectrum='#',
                                               h_subset_filtering=0,
                                               min_occ=2,
                                               k_superset_filtering=0,
                                               l_covered_spikes=0)
        self.assertEqual(concepts, [concept1])

        # reject concept1 using superset filtering
        # (case with non-empty intersection but no superset)
        concepts = spade.pattern_set_reduction([concept1, concept3],
                                               ns_signatures=[], min_spikes=2,
                                               winlen=winlen, spectrum='#',
                                               k_superset_filtering=0)
        self.assertEqual(concepts, [concept3])

        # keep concept1 by increasing k_superset_filtering
        concepts = spade.pattern_set_reduction([concept1, concept3],
                                               ns_signatures=[], min_spikes=2,
                                               winlen=winlen, spectrum='#',
                                               k_superset_filtering=1)
        self.assertEqual(concepts, [concept1, concept3])

        # reject concept3 using ns_signatures
        concepts = spade.pattern_set_reduction([concept1, concept3],
                                               ns_signatures=[(3, 2)],
                                               min_spikes=2,
                                               winlen=winlen, spectrum='#',
                                               k_superset_filtering=1)
        self.assertEqual(concepts, [concept1])

        # reject concept3 using the covered spikes criterion
        concepts = spade.pattern_set_reduction([concept1, concept3],
                                               ns_signatures=[(3, 2), (2, 3)],
                                               min_spikes=2,
                                               winlen=winlen, spectrum='#',
                                               k_superset_filtering=1,
                                               l_covered_spikes=0)
        self.assertEqual(concepts, [concept1])

        # check that two concepts with disjoint intents are both kept
        concepts = spade.pattern_set_reduction([concept3, concept4],
                                               ns_signatures=[],
                                               winlen=winlen, spectrum='#')
        self.assertEqual(concepts, [concept3, concept4])

    @unittest.skipUnless(HAVE_STATSMODELS,
                         "'fdr_bh' stat corr requires statsmodels")
    def test_signature_significance_fdr_bh_corr(self):
        """
        A typical corr='fdr_bh' scenario, that requires statsmodels.
        """
        sig_spectrum = spade.test_signature_significance(
            pv_spec=((2, 3, 0.2), (2, 4, 0.05)),
            concepts=([[(2, 3), (1, 2, 3)],
                       [(2, 4), (1, 2, 3, 4)]]),
            alpha=0.15, winlen=1, corr='fdr_bh')
        self.assertEqual(sig_spectrum, [(2., 3., False), (2., 4., True)])



# -*- coding: utf-8 -*-
"""
Tests for the spike train dissimilarity measures module.

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""
import unittest
from neo import SpikeTrain
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
import scipy.integrate as spint
from quantities import ms, s, Hz
import elephant.kernels as kernels
from elephant.spike_train_generation import StationaryPoissonProcess
import elephant.spike_train_dissimilarity as stds

from elephant.datasets import download_datasets


class TimeScaleDependSpikeTrainDissimMeasuresTestCase(unittest.TestCase):
    def setUp(self):
        self.st00 = SpikeTrain([], units='ms', t_stop=1000.0)
        self.st01 = SpikeTrain([1], units='ms', t_stop=1000.0)
        self.st02 = SpikeTrain([2], units='ms', t_stop=1000.0)
        self.st03 = SpikeTrain([2.9], units='ms', t_stop=1000.0)
        self.st04 = SpikeTrain([3.1], units='ms', t_stop=1000.0)
        self.st05 = SpikeTrain([5], units='ms', t_stop=1000.0)
        self.st06 = SpikeTrain([500], units='ms', t_stop=1000.0)
        self.st07 = SpikeTrain([12, 32], units='ms', t_stop=1000.0)
        self.st08 = SpikeTrain([32, 52], units='ms', t_stop=1000.0)
        self.st09 = SpikeTrain([42], units='ms', t_stop=1000.0)
        self.st10 = SpikeTrain([18, 60], units='ms', t_stop=1000.0)
        self.st11 = SpikeTrain([10, 20, 30, 40], units='ms', t_stop=1000.0)
        self.st12 = SpikeTrain([40, 30, 20, 10], units='ms', t_stop=1000.0)
        self.st13 = SpikeTrain([15, 25, 35, 45], units='ms', t_stop=1000.0)
        self.st14 = SpikeTrain([10, 20, 30, 40, 50], units='ms', t_stop=1000.0)
        self.st15 = SpikeTrain([0.01, 0.02, 0.03, 0.04, 0.05],
                               units='s', t_stop=1000.0)
        self.st16 = SpikeTrain([12, 16, 28, 30, 42], units='ms', t_stop=1000.0)
        self.st21 = StationaryPoissonProcess(rate=50 * Hz, t_start=0 * ms,
                                             t_stop=1000 * ms
                                             ).generate_spiketrain()
        self.st22 = StationaryPoissonProcess(rate=40 * Hz, t_start=0 * ms,
                                             t_stop=1000 * ms
                                             ).generate_spiketrain()
        self.st23 = StationaryPoissonProcess(rate=30 * Hz, t_start=0 * ms,
                                             t_stop=1000 * ms
                                             ).generate_spiketrain()
        self.rd_st_list = [self.st21, self.st22, self.st23]
        self.st31 = SpikeTrain([12.0], units='ms', t_stop=1000.0)
        self.st32 = SpikeTrain([12.0, 12.0], units='ms', t_stop=1000.0)
        self.st33 = SpikeTrain([20.0], units='ms', t_stop=1000.0)
        self.st34 = SpikeTrain([20.0, 20.0], units='ms', t_stop=1000.0)
        self.array1 = np.arange(1, 10)
        self.array2 = np.arange(1.2, 10)
        self.qarray1 = self.array1 * Hz
        self.qarray2 = self.array2 * Hz
        self.tau0 = 0.0 * ms
        self.q0 = np.inf / ms
        self.tau1 = 0.000000001 * ms
        self.q1 = 1.0 / self.tau1
        self.tau2 = 1.0 * ms
        self.q2 = 1.0 / self.tau2
        self.tau3 = 10.0 * ms
        self.q3 = 1.0 / self.tau3
        self.tau4 = 100.0 * ms
        self.q4 = 1.0 / self.tau4
        self.tau5 = 1000000000.0 * ms
        self.q5 = 1.0 / self.tau5
        self.tau6 = np.inf * ms
        self.q6 = 0.0 / ms
        self.tau7 = 0.01 * s
        self.q7 = 1.0 / self.tau7
        self.t = np.linspace(0, 200, 20000001) * ms

    def test_wrong_input(self):
        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.array1, self.array2], self.q3)
        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.qarray1, self.qarray2], self.q3)
        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.qarray1, self.qarray2], 5.0 * ms)

        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.array1, self.array2], self.q3,
                          algorithm='intuitive')
        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.qarray1, self.qarray2], self.q3,
                          algorithm='intuitive')
        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.qarray1, self.qarray2], 5.0 * ms,
                          algorithm='intuitive')

        self.assertRaises(TypeError, stds.van_rossum_distance,
                          [self.array1, self.array2], self.tau3)
        self.assertRaises(TypeError, stds.van_rossum_distance,
                          [self.qarray1, self.qarray2], self.tau3)
        self.assertRaises(TypeError, stds.van_rossum_distance,
                          [self.qarray1, self.qarray2], 5.0 * Hz)

        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.st11, self.st13], self.tau2)
        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.st11, self.st13], 5.0)
        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.st11, self.st13], self.tau2,
                          algorithm='intuitive')
        self.assertRaises(TypeError, stds.victor_purpura_distance,
                          [self.st11, self.st13], 5.0,
                          algorithm='intuitive')
        self.assertRaises(TypeError, stds.van_rossum_distance,
                          [self.st11, self.st13], self.q4)
        self.assertRaises(TypeError, stds.van_rossum_distance,
                          [self.st11, self.st13], 5.0)

        self.assertRaises(NotImplementedError, stds.victor_purpura_distance,
                          [self.st01, self.st02], self.q3,
                          kernel=kernels.Kernel(2.0 / self.q3))
        self.assertRaises(NotImplementedError, stds.victor_purpura_distance,
                          [self.st01, self.st02], self.q3,
                          kernel=kernels.SymmetricKernel(2.0 / self.q3))
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st02], self.q1,
            kernel=kernels.TriangularKernel(
                2.0 / (np.sqrt(6.0) * self.q2)))[0, 1],
                         stds.victor_purpura_distance(
                             [self.st01, self.st02], self.q3,
                             kernel=kernels.TriangularKernel(
                                 2.0 / (np.sqrt(6.0) * self.q2)))[0, 1])
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st02],
            kernel=kernels.TriangularKernel(
                2.0 / (np.sqrt(6.0) * self.q2)))[0, 1], 1.0)
        self.assertNotEqual(stds.victor_purpura_distance(
            [self.st01, self.st02],
            kernel=kernels.AlphaKernel(
                2.0 / (np.sqrt(6.0) * self.q2)))[0, 1], 1.0)

        self.assertRaises(NameError, stds.victor_purpura_distance,
                          [self.st11, self.st13], self.q2, algorithm='slow')

    def test_victor_purpura_distance_fast(self):
        # Tests of distances of simplest spike trains:
        self.assertEqual(stds.victor_purpura_distance(
            [self.st00, self.st00], self.q2)[0, 1], 0.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st00, self.st01], self.q2)[0, 1], 1.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st00], self.q2)[0, 1], 1.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st01], self.q2)[0, 1], 0.0)
        # Tests of distances under elementary spike operations
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st02], self.q2)[0, 1], 1.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st03], self.q2)[0, 1], 1.9)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st04], self.q2)[0, 1], 2.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st05], self.q2)[0, 1], 2.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st00, self.st07], self.q2)[0, 1], 2.0)
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st07, self.st08], self.q4)[0, 1], 0.4)
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st07, self.st10], self.q3)[0, 1], 0.6 + 2)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st11, self.st14], self.q2)[0, 1], 1)
        # Tests on timescales
        self.assertEqual(stds.victor_purpura_distance(
            [self.st11, self.st14], self.q1)[0, 1],
                         stds.victor_purpura_distance(
                             [self.st11, self.st14], self.q5)[0, 1])
        self.assertEqual(stds.victor_purpura_distance(
            [self.st07, self.st11], self.q0)[0, 1], 6.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st07, self.st11], self.q1)[0, 1], 6.0)
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st07, self.st11], self.q5)[0, 1], 2.0, 5)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st07, self.st11], self.q6)[0, 1], 2.0)
        # Tests on unordered spiketrains
        self.assertEqual(stds.victor_purpura_distance(
            [self.st11, self.st13], self.q4)[0, 1],
                         stds.victor_purpura_distance(
                             [self.st12, self.st13], self.q4)[0, 1])
        self.assertNotEqual(stds.victor_purpura_distance(
            [self.st11, self.st13], self.q4,
            sort=False)[0, 1],
                            stds.victor_purpura_distance(
                                [self.st12, self.st13], self.q4,
                                sort=False)[0, 1])
        # Tests on metric properties with random spiketrains
        # (explicit calculation of second metric axiom in particular case,
        # because from dist_matrix it is trivial)
        dist_matrix = stds.victor_purpura_distance(
            [self.st21, self.st22, self.st23], self.q3)
        for i in range(3):
            for j in range(3):
                self.assertGreaterEqual(dist_matrix[i, j], 0)
                if dist_matrix[i, j] == 0:
                    assert_array_equal(self.rd_st_list[i], self.rd_st_list[j])
        assert_array_equal(stds.victor_purpura_distance(
            [self.st21, self.st22], self.q3),
            stds.victor_purpura_distance(
                [self.st22, self.st21], self.q3))
        self.assertLessEqual(dist_matrix[0, 1],
                             dist_matrix[0, 2] + dist_matrix[1, 2])
        self.assertLessEqual(dist_matrix[0, 2],
                             dist_matrix[1, 2] + dist_matrix[0, 1])
        self.assertLessEqual(dist_matrix[1, 2],
                             dist_matrix[0, 1] + dist_matrix[0, 2])
        # Tests on proper unit conversion
        self.assertAlmostEqual(
            stds.victor_purpura_distance([self.st14, self.st16],
                                         self.q3)[0, 1],
            stds.victor_purpura_distance([self.st15, self.st16],
                                         self.q3)[0, 1])
        self.assertAlmostEqual(
            stds.victor_purpura_distance([self.st16, self.st14],
                                         self.q3)[0, 1],
            stds.victor_purpura_distance([self.st16, self.st15],
                                         self.q3)[0, 1])
        self.assertAlmostEqual(
            stds.victor_purpura_distance([self.st01, self.st05],
                                         self.q3)[0, 1],
            stds.victor_purpura_distance([self.st01, self.st05],
                                         self.q7)[0, 1])
        # Tests on algorithmic behaviour for equal spike times
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st31, self.st34], self.q3)[0, 1], 0.8 + 1.0)
        self.assertAlmostEqual(
            stds.victor_purpura_distance([self.st31, self.st34],
                                         self.q3)[0, 1],
            stds.victor_purpura_distance([self.st32, self.st33],
                                         self.q3)[0, 1])
        self.assertAlmostEqual(
            stds.victor_purpura_distance(
                [self.st31, self.st33], self.q3)[0, 1] * 2.0,
            stds.victor_purpura_distance(
                [self.st32, self.st34], self.q3)[0, 1])
        # Tests on spike train list lengthes smaller than 2
        self.assertEqual(stds.victor_purpura_distance(
            [self.st21], self.q3)[0, 0], 0)
        self.assertEqual(len(stds.victor_purpura_distance([], self.q3)), 0)

    def test_victor_purpura_distance_intuitive(self):
        # Tests of distances of simplest spike trains
        self.assertEqual(stds.victor_purpura_distance(
            [self.st00, self.st00], self.q2,
            algorithm='intuitive')[0, 1], 0.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st00, self.st01], self.q2,
            algorithm='intuitive')[0, 1], 1.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st00], self.q2,
            algorithm='intuitive')[0, 1], 1.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st01], self.q2,
            algorithm='intuitive')[0, 1], 0.0)
        # Tests of distances under elementary spike operations
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st02], self.q2,
            algorithm='intuitive')[0, 1], 1.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st03], self.q2,
            algorithm='intuitive')[0, 1], 1.9)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st04], self.q2,
            algorithm='intuitive')[0, 1], 2.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st05], self.q2,
            algorithm='intuitive')[0, 1], 2.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st00, self.st07], self.q2,
            algorithm='intuitive')[0, 1], 2.0)
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st07, self.st08], self.q4,
            algorithm='intuitive')[0, 1], 0.4)
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st07, self.st10], self.q3,
            algorithm='intuitive')[0, 1], 2.6)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st11, self.st14], self.q2,
            algorithm='intuitive')[0, 1], 1)
        # Tests on timescales
        self.assertEqual(stds.victor_purpura_distance(
            [self.st11, self.st14], self.q1,
            algorithm='intuitive')[0, 1],
                         stds.victor_purpura_distance(
                             [self.st11, self.st14], self.q5,
                             algorithm='intuitive')[0, 1])
        self.assertEqual(stds.victor_purpura_distance(
            [self.st07, self.st11], self.q0,
            algorithm='intuitive')[0, 1], 6.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st07, self.st11], self.q1,
            algorithm='intuitive')[0, 1], 6.0)
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st07, self.st11], self.q5,
            algorithm='intuitive')[0, 1], 2.0, 5)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st07, self.st11], self.q6,
            algorithm='intuitive')[0, 1], 2.0)
        # Tests on unordered spiketrains
        self.assertEqual(stds.victor_purpura_distance(
            [self.st11, self.st13], self.q4,
            algorithm='intuitive')[0, 1],
                         stds.victor_purpura_distance(
                             [self.st12, self.st13], self.q4,
                             algorithm='intuitive')[0, 1])
        self.assertNotEqual(stds.victor_purpura_distance(
            [self.st11, self.st13], self.q4,
            sort=False, algorithm='intuitive')[0, 1],
                            stds.victor_purpura_distance(
                                [self.st12, self.st13], self.q4,
                                sort=False, algorithm='intuitive')[0, 1])
        # Tests on metric properties with random spiketrains
        # (explicit calculation of second metric axiom in particular case,
        # because from dist_matrix it is trivial)
        dist_matrix = stds.victor_purpura_distance(
            [self.st21, self.st22, self.st23],
            self.q3, algorithm='intuitive')
        for i in range(3):
            for j in range(3):
                self.assertGreaterEqual(dist_matrix[i, j], 0)
                if dist_matrix[i, j] == 0:
                    assert_array_equal(self.rd_st_list[i], self.rd_st_list[j])
        assert_array_equal(stds.victor_purpura_distance(
            [self.st21, self.st22], self.q3,
            algorithm='intuitive'),
            stds.victor_purpura_distance(
                [self.st22, self.st21], self.q3,
                algorithm='intuitive'))
        self.assertLessEqual(dist_matrix[0, 1],
                             dist_matrix[0, 2] + dist_matrix[1, 2])
        self.assertLessEqual(dist_matrix[0, 2],
                             dist_matrix[1, 2] + dist_matrix[0, 1])
        self.assertLessEqual(dist_matrix[1, 2],
                             dist_matrix[0, 1] + dist_matrix[0, 2])
        # Tests on proper unit conversion
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st14, self.st16], self.q3,
            algorithm='intuitive')[0, 1],
                               stds.victor_purpura_distance(
                                   [self.st15, self.st16], self.q3,
                                   algorithm='intuitive')[0, 1])
        self.assertAlmostEqual(stds.victor_purpura_distance(
            [self.st16, self.st14], self.q3,
            algorithm='intuitive')[0, 1],
                               stds.victor_purpura_distance(
                                   [self.st16, self.st15], self.q3,
                                   algorithm='intuitive')[0, 1])
        self.assertEqual(stds.victor_purpura_distance(
            [self.st01, self.st05], self.q3,
            algorithm='intuitive')[0, 1],
                         stds.victor_purpura_distance(
                             [self.st01, self.st05], self.q7,
                             algorithm='intuitive')[0, 1])
        # Tests on algorithmic behaviour for equal spike times
        self.assertEqual(stds.victor_purpura_distance(
            [self.st31, self.st34], self.q3,
            algorithm='intuitive')[0, 1],
                         0.8 + 1.0)
        self.assertEqual(stds.victor_purpura_distance(
            [self.st31, self.st34], self.q3,
            algorithm='intuitive')[0, 1],
                         stds.victor_purpura_distance(
                             [self.st32, self.st33], self.q3,
                             algorithm='intuitive')[0, 1])
        self.assertEqual(stds.victor_purpura_distance(
            [self.st31, self.st33], self.q3,
            algorithm='intuitive')[0, 1] * 2.0,
                         stds.victor_purpura_distance(
                             [self.st32, self.st34], self.q3,
                             algorithm='intuitive')[0, 1])
        # Tests on spike train list lengthes smaller than 2
        self.assertEqual(stds.victor_purpura_distance(
            [self.st21], self.q3,
            algorithm='intuitive')[0, 0], 0)
        self.assertEqual(len(stds.victor_purpura_distance(
            [], self.q3, algorithm='intuitive')), 0)

    def test_victor_purpura_algorithm_comparison(self):
        assert_array_almost_equal(
            stds.victor_purpura_distance([self.st21, self.st22, self.st23],
                                         self.q3),
            stds.victor_purpura_distance([self.st21, self.st22, self.st23],
                                         self.q3, algorithm='intuitive'))

    def test_victor_purpura_matlab_comparison_float(self):

        repo_path =\
            r"unittest/spike_train_dissimilarity/victor_purpura_distance/data"

        files_to_download = [
            ("times_float.npy", "ed1ff4d2c0eeed4a2b50a456803656be"),
            ("matlab_results_float.npy", "a17f049e7ad0ddf7ca812e86fdb92646")]

        downloaded_files = {}
        for filename, checksum in files_to_download:
            downloaded_files[filename] = {
                'filename': filename,
                'path': download_datasets(repo_path=f"{repo_path}/{filename}",
                                          checksum=checksum)}

        times_float = np.load(downloaded_files['times_float.npy']['path'])
        mat_res_float = np.load(downloaded_files[
            'matlab_results_float.npy']['path'])

        r_float = SpikeTrain(times_float[0], units='ms', t_start=0,
                             t_stop=1000 * ms)
        s_float = SpikeTrain(times_float[1], units='ms', t_start=0,
                             t_stop=1000 * ms)
        t_float = SpikeTrain(times_float[2], units='ms', t_start=0,
                             t_stop=1000 * ms)

        vic_pur_result_float = stds.victor_purpura_distance(
            [r_float, s_float, t_float],
            cost_factor=1.0 / ms, kernel=None,
            sort=True, algorithm='intuitive')

        assert_array_almost_equal(vic_pur_result_float, mat_res_float)

    def test_victor_purpura_matlab_comparison_int(self):

        repo_path =\
            r"unittest/spike_train_dissimilarity/victor_purpura_distance/data"

        files_to_download = [
            ("times_int.npy", "aa1411c04da3f58d8b8913ae2f935057"),
            ("matlab_results_int.npy", "7edd32e50edde12dc1ef4aa5f57f70fb")]

        downloaded_files = {}
        for filename, checksum in files_to_download:
            downloaded_files[filename] = {
                'filename': filename,
                'path': download_datasets(repo_path=f"{repo_path}/{filename}",
                                          checksum=checksum)}

        times_int = np.load(downloaded_files['times_int.npy']['path'])
        mat_res_int = np.load(
            downloaded_files['matlab_results_int.npy']['path'])

        r_int = SpikeTrain(times_int[0], units='ms', t_start=0,
                           t_stop=1000 * ms)
        s_int = SpikeTrain(times_int[1], units='ms', t_start=0,
                           t_stop=1000 * ms)
        t_int = SpikeTrain(times_int[2], units='ms', t_start=0,
                           t_stop=1000 * ms)

        vic_pur_result_int = stds.victor_purpura_distance(
            [r_int, s_int, t_int],
            cost_factor=1.0 / ms, kernel=None,
            sort=True, algorithm='intuitive')

        assert_array_equal(vic_pur_result_int, mat_res_int)

    def test_van_rossum_distance(self):
        # Tests of distances of simplest spike trains
        self.assertEqual(stds.van_rossum_distance(
            [self.st00, self.st00], self.tau2)[0, 1], 0.0)
        self.assertEqual(stds.van_rossum_distance(
            [self.st00, self.st01], self.tau2)[0, 1], 1.0)
        self.assertEqual(stds.van_rossum_distance(
            [self.st01, self.st00], self.tau2)[0, 1], 1.0)
        self.assertEqual(stds.van_rossum_distance(
            [self.st01, self.st01], self.tau2)[0, 1], 0.0)
        # Tests of distances under elementary spike operations
        self.assertAlmostEqual(stds.van_rossum_distance(
            [self.st01, self.st02], self.tau2)[0, 1],
                               float(np.sqrt(2 * (1.0 - np.exp(-np.absolute(
                                   ((self.st01[0] - self.st02[0]) /
                                    self.tau2).simplified))))))
        self.assertAlmostEqual(stds.van_rossum_distance(
            [self.st01, self.st05], self.tau2)[0, 1],
                               float(np.sqrt(2 * (1.0 - np.exp(-np.absolute(
                                   ((self.st01[0] - self.st05[0]) /
                                    self.tau2).simplified))))))
        self.assertAlmostEqual(stds.van_rossum_distance(
            [self.st01, self.st05], self.tau2)[0, 1],
                               np.sqrt(2.0), 1)
        self.assertAlmostEqual(stds.van_rossum_distance(
            [self.st01, self.st06], self.tau2)[0, 1],
                               np.sqrt(2.0), 20)
        self.assertAlmostEqual(stds.van_rossum_distance(
            [self.st00, self.st07], self.tau1)[0, 1],
                               np.sqrt(0 + 2))
        self.assertAlmostEqual(stds.van_rossum_distance(
            [self.st07, self.st08], self.tau4)[0, 1],
                               float(np.sqrt(2 * (1.0 - np.exp(-np.absolute(
                                   ((self.st07[0] - self.st08[-1]) /
                                    self.tau4).simplified))))))
        f_minus_g_squared = (
            (self.t > self.st08[0]) * np.exp(
                -((self.t - self.st08[0]) / self.tau3).simplified) +
            (self.t > self.st08[1]) * np.exp(
                -((self.t - self.st08[1]) / self.tau3).simplified) -
            (self.t > self.st09[0]) * np.exp(
                -((self.t - self.st09[0]) / self.tau3).simplified)) ** 2
        distance = np.sqrt(2.0 * spint.cumulative_trapezoid(
            y=f_minus_g_squared, x=self.t.magnitude)[-1] /
                           self.tau3.rescale(self.t.units).magnitude)
        self.assertAlmostEqual(stds.van_rossum_distance(
            [self.st08, self.st09], self.tau3)[0, 1], distance, 5)
        self.assertAlmostEqual(stds.van_rossum_distance(
            [self.st11, self.st14], self.tau2)[0, 1], 1)
        # Tests on timescales
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st11, self.st14], self.tau1)[0, 1],
            stds.van_rossum_distance([self.st11, self.st14], self.tau5)[0, 1])

        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st07, self.st11], self.tau0)[0, 1],
            np.sqrt(len(self.st07) + len(self.st11)))
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st07, self.st14], self.tau0)[0, 1],
            np.sqrt(len(self.st07) + len(self.st14)))
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st07, self.st11], self.tau1)[0, 1],
            np.sqrt(len(self.st07) + len(self.st11)))
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st07, self.st14], self.tau1)[0, 1],
            np.sqrt(len(self.st07) + len(self.st14)))
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st07, self.st11], self.tau5)[0, 1],
            np.absolute(len(self.st07) - len(self.st11)))
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st07, self.st14], self.tau5)[0, 1],
            np.absolute(len(self.st07) - len(self.st14)))
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st07, self.st11], self.tau6)[0, 1],
            np.absolute(len(self.st07) - len(self.st11)))
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st07, self.st14], self.tau6)[0, 1],
            np.absolute(len(self.st07) - len(self.st14)))
        # Tests on unordered spiketrains
        self.assertEqual(
            stds.van_rossum_distance([self.st11, self.st13], self.tau4)[0, 1],
            stds.van_rossum_distance([self.st12, self.st13], self.tau4)[0, 1])
        self.assertNotEqual(
            stds.van_rossum_distance([self.st11, self.st13],
                                     self.tau4, sort=False)[0, 1],
            stds.van_rossum_distance([self.st12, self.st13],
                                     self.tau4, sort=False)[0, 1])
        # Tests on metric properties with random spiketrains
        # (explicit calculation of second metric axiom in particular case,
        # because from dist_matrix it is trivial)
        dist_matrix = stds.van_rossum_distance(
            [self.st21, self.st22, self.st23], self.tau3)
        for i in range(3):
            for j in range(3):
                self.assertGreaterEqual(dist_matrix[i, j], 0)
                if dist_matrix[i, j] == 0:
                    assert_array_equal(self.rd_st_list[i], self.rd_st_list[j])
        assert_array_equal(
            stds.van_rossum_distance([self.st21, self.st22], self.tau3),
            stds.van_rossum_distance([self.st22, self.st21], self.tau3))
        self.assertLessEqual(dist_matrix[0, 1],
                             dist_matrix[0, 2] + dist_matrix[1, 2])
        self.assertLessEqual(dist_matrix[0, 2],
                             dist_matrix[1, 2] + dist_matrix[0, 1])
        self.assertLessEqual(dist_matrix[1, 2],
                             dist_matrix[0, 1] + dist_matrix[0, 2])
        # Tests on proper unit conversion
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st14, self.st16], self.tau3)[0, 1],
            stds.van_rossum_distance([self.st15, self.st16], self.tau3)[0, 1])
        self.assertAlmostEqual(
            stds.van_rossum_distance([self.st16, self.st14], self.tau3)[0, 1],
            stds.van_rossum_distance([self.st16, self.st15], self.tau3)[0, 1])
        self.assertEqual(
            stds.van_rossum_distance([self.st01, self.st05], self.tau3)[0, 1],
            stds.van_rossum_distance([self.st01, self.st05], self.tau7)[0, 1])
        # Tests on algorithmic behaviour for equal spike times
        f_minus_g_squared = (
              (self.t > self.st31[0]) * np.exp(
                -((self.t - self.st31[0]) / self.tau3).simplified) -
              (self.t > self.st34[0]) * np.exp(
                -((self.t - self.st34[0]) / self.tau3).simplified) -
              (self.t > self.st34[1]) * np.exp(
                -((self.t - self.st34[1]) / self.tau3).simplified)) ** 2
        distance = np.sqrt(2.0 * spint.cumulative_trapezoid(
            y=f_minus_g_squared, x=self.t.magnitude)[-1] /
                           self.tau3.rescale(self.t.units).magnitude)
        self.assertAlmostEqual(stds.van_rossum_distance([self.st31, self.st34],
                                                        self.tau3)[0, 1],
                               distance, 5)
        self.assertEqual(stds.van_rossum_distance([self.st31, self.st34],
                                                  self.tau3)[0, 1],
                         stds.van_rossum_distance([self.st32, self.st33],
                                                  self.tau3)[0, 1])
        self.assertEqual(stds.van_rossum_distance([self.st31, self.st33],
                                                  self.tau3)[0, 1] * 2.0,
                         stds.van_rossum_distance([self.st32, self.st34],
                                                  self.tau3)[0, 1])
        # Tests on spike train list lengthes smaller than 2
        self.assertEqual(stds.van_rossum_distance(
            [self.st21], self.tau3)[0, 0], 0)
        self.assertEqual(len(stds.van_rossum_distance([], self.tau3)), 0)


if __name__ == '__main__':
    unittest.main()
